import numpy as np
from nleis.nleis import EISandNLEIS, NLEISCustomCircuit
import os

import os.path
import sys  # noqa: F401


def test_model_io():

    test_dir = os.path.dirname(os.path.abspath(__file__))
    data_dir = os.path.join(test_dir, '../data')

    # # get example data
    # # The example is shown in "Getting Started" page

    frequencies = np.loadtxt(os.path.join(data_dir, 'freq_30a.txt'))
    Z1 = np.loadtxt(os.path.join(data_dir, 'Z1s_30a.txt')).view(complex)[1]
    Z2 = np.loadtxt(os.path.join(data_dir, 'Z2s_30a.txt')).view(complex)[1]
    circ_str_1 = 'L0-R0-TDS0-TDS1'
    circ_str_2 = 'd(TDSn0,TDSn1)'

    # test for EISandNLEIS
    initial_guess = [1e-7, 1e-3,  # L0,RO
                     5e-3, 1e-3, 10, 1e-2, 100, 10, 0.1,
                     # TDS0 + additioal nonlinear parameters
                     1e-3, 1e-3, 1e-3, 1e-2, 1000, 0, 0
                     # TDS1 + additioal nonlinear parameters
                     ]

    circuit_1 = EISandNLEIS(circ_str_1, circ_str_2,
                            initial_guess=initial_guess)

    # circuit_1.save('./nleis_tests/test_io.json')
    circuit_1.save(os.path.join(data_dir, 'test_io.json'))

    circuit_2 = EISandNLEIS()
    circuit_2.load(os.path.join(data_dir, 'test_io.json'))

    assert circuit_1 == circuit_2

    circuit_1.fit(frequencies, Z1, Z2)
    p_fit = list(circuit_1.parameters_)
    circuit_1.save(os.path.join(data_dir, 'test_io.json'))
    circuit_2 = EISandNLEIS()
    circuit_2.load(os.path.join(data_dir, 'test_io.json'))

    assert str(circuit_1) == str(circuit_2)
    assert circuit_1 == circuit_2

    fitted_template = EISandNLEIS()
    fitted_template.load(os.path.join(
        data_dir, 'test_io.json'), fitted_as_initial=True)
    circuit_1 = EISandNLEIS(circ_str_1, circ_str_2, initial_guess=p_fit)
    assert str(circuit_1) == str(fitted_template)
    assert circuit_1 == fitted_template

    # test for NLIESCustomCircuit
    initial_guess = [
        5e-3, 1e-3, 10, 1e-2, 100, 10, 0.1,
        # TDS0 + additioal nonlinear parameters
        1e-3, 1e-3, 1e-3, 1e-2, 1000, 0, 0
        # TDS1 + additioal nonlinear parameters
    ]

    circuit_1 = NLEISCustomCircuit(circ_str_2,
                                   initial_guess=initial_guess)

    circuit_1.save(os.path.join(data_dir, 'test_io.json'))

    circuit_2 = NLEISCustomCircuit()
    circuit_2.load(os.path.join(data_dir, 'test_io.json'))

    assert circuit_1 == circuit_2

    circuit_1.fit(frequencies, Z2)
    p_fit = list(circuit_1.parameters_)
    circuit_1.save(os.path.join(data_dir, 'test_io.json'))
    circuit_2 = NLEISCustomCircuit()
    circuit_2.load(os.path.join(data_dir, 'test_io.json'))

    assert str(circuit_1) == str(circuit_2)
    assert circuit_1 == circuit_2

    fitted_template = NLEISCustomCircuit()
    fitted_template.load(os.path.join(
        data_dir, 'test_io.json'), fitted_as_initial=True)
    circuit_1 = NLEISCustomCircuit(circ_str_2, initial_guess=p_fit)



import json  # noqa: F401
import os

import numpy as np
import matplotlib.pyplot as plt  # noqa: F401
import pytest

from nleis.nleis import EISandNLEIS, NLEISCustomCircuit  # noqa: F401

from nleis.nleis_fitting import data_processing

test_dir = os.path.dirname(os.path.abspath(__file__))
data_dir = os.path.join(test_dir, '../data')

# # get example data
# # The example is shown in "Getting Started" page

frequencies = np.loadtxt(os.path.join(data_dir, 'freq_30a.txt'))
Z1 = np.loadtxt(os.path.join(data_dir, 'Z1s_30a.txt')).view(complex)[1]
Z2 = np.loadtxt(os.path.join(data_dir, 'Z2s_30a.txt')).view(complex)[1]

f, Z1, Z2, f2_trunc, Z2_trunc = data_processing(frequencies, Z1, Z2)


def test_parsing():
    circ_str_1 = 'L0-R0-TDS0-TDS1'
    circ_str_2 = 'd(TDSn0,TDSn1)'
    initial_guess = [1e-7, 1e-3,  # L0,RO
                     5e-3, 1e-3, 1e-2, 100, 10, 0.1,
                     # TDS0 + additioal nonlinear parameters
                     1e-3, 1e-3, 1e-3, 1e-2, 1000, 0,
                     # TDS1 + additioal nonlinear parameters
                     ]

    NLEIS_circuit = EISandNLEIS(
        circ_str_1, circ_str_2, initial_guess=initial_guess,
        constants={'TDSn0_2': 10, 'TDSn1_6': 0})

    assert NLEIS_circuit.edited_circuit == 'L0-R0-TDSn0-TDSn1'
    assert NLEIS_circuit.constants_1 == {'TDS0_2': 10}
    assert NLEIS_circuit.constants_2 == {'TDSn0_2': 10, 'TDSn1_6': 0}


def test_EISandNLEIS():
    circ_str_1 = 'L0-R0-TDS0-TDS1'
    circ_str_2 = 'd(TDSn0,TDSn1)'
    initial_guess = [1e-7, 1e-3,  # L0,RO
                     5e-3, 1e-3, 10, 1e-2, 100, 10, 0.1,
                     # TDS0 + additioal nonlinear parameters
                     1e-3, 1e-3, 1e-3, 1e-2, 1000, 0, 0,
                     # TDS1 + additioal nonlinear parameters
                     ]

    NLEIS_circuit = EISandNLEIS(
        circ_str_1, circ_str_2, initial_guess=initial_guess)

    # check get_param_names()
    full_names_EIS, all_units_EIS = NLEIS_circuit.get_param_names(
        circ_str_1, {})

    assert full_names_EIS == ['L0', 'R0', 'TDS0_0', 'TDS0_1', 'TDS0_2',
                              'TDS0_3', 'TDS0_4',
                              'TDS1_0', 'TDS1_1', 'TDS1_2', 'TDS1_3', 'TDS1_4']
    assert all_units_EIS == ['H', 'Ohm', 'Ohms', 'Ohms',
                             'F', 'Ohms', 's', 'Ohms', 'Ohms',
                             'F', 'Ohms', 's']
    full_names_NLEIS, all_units_NLEIS = NLEIS_circuit.get_param_names(
        circ_str_2, {})
    assert full_names_NLEIS == ['TDSn0_0', 'TDSn0_1', 'TDSn0_2', 'TDSn0_3',
                                'TDSn0_4', 'TDSn0_5',
                                'TDSn0_6', 'TDSn1_0', 'TDSn1_1', 'TDSn1_2',
                                'TDSn1_3', 'TDSn1_4', 'TDSn1_5', 'TDSn1_6']
    assert all_units_NLEIS == ['Ohms', 'Ohms', 'F', 'Ohms', 's',
                               '1/V', '-', 'Ohms', 'Ohms', 'F', 'Ohms', 's',
                               '1/V', '-']

    # check _is_fit()
    assert not NLEIS_circuit._is_fit()

    # check complex frequencies raise TypeError
    with pytest.raises(TypeError):
        NLEIS_circuit.predict([0.42, 42 + 42j])

    initial_guess = [1e-3,  # RO
                     5e-3, 1e-3, 10, 1e-2, 100, 10, 0.1,
                     # TDS0 + additioal nonlinear parameters
                     1e-3, 1e-3, 1e-3, 1e-2, 1000, 0,
                     # TDS1 + additioal nonlinear parameters
                     ]
    # check correct assignment of constants
    # when EIS element is supplied
    NLEIS_circuit = EISandNLEIS(
        circ_str_1, circ_str_2, initial_guess=initial_guess,
        constants={'L0': 1e-7, 'TDS1_3': 1})

    assert {'L0': 1e-7, 'TDS1_3': 1} == NLEIS_circuit.constants_1
    assert {'TDSn1_3': 1} == NLEIS_circuit.constants_2

    # check correct assignment of constants
    # when 2nd-NLEIS element is supplied
    NLEIS_circuit = EISandNLEIS(
        circ_str_1, circ_str_2, initial_guess=initial_guess,
        constants={'L0': 1e-7, 'TDSn1_6': 0})
    assert {'L0': 1e-7} == NLEIS_circuit.constants_1
    assert {'TDSn1_6': 0} == NLEIS_circuit.constants_2

    # raise ValueError if the length of the frequency
    # does not matches with the length of the impedance for EIS
    with pytest.raises(ValueError):
        NLEIS_circuit.fit(f[f < 1], Z1, Z2)

    # raise ValueError if no initial_guess is supplied
    with pytest.raises(ValueError):
        NLEIS_circuit = EISandNLEIS(
            circ_str_1, circ_str_2, initial_guess=initial_guess,
            constants={'L0': 1e-7, 'TDSn1_6': 0})
        NLEIS_circuit.fit(f[f < 1], Z1, Z2)

    # check non-number initial guess raise TypeError
    with pytest.raises(TypeError):
        NLEIS_circuit = EISandNLEIS(
            circ_str_1, circ_str_2, initial_guess=['a'],
            constants={'L0': 1e-7, 'TDSn1_6': 0})

    # check mismatched EIS and 2nd-NLEIS circuit
    # raise TypeError
    with pytest.raises(TypeError):
        NLEIS_circuit = EISandNLEIS(
            circ_str_1, 'd(TDS0,TDSn2)', initial_guess=initial_guess,
            constants={'L0': 1e-7, 'TDSn1_6': 0})

    # check wrong constants raise
    # raise ValueError
    with pytest.raises(ValueError):
        NLEIS_circuit = EISandNLEIS(
            circ_str_1, circ_str_2, initial_guess=initial_guess,
            constants={'L0': 1e-7, 'TDSn2_6': 0})

    # check that constants number beyond the range
    # of allowed number of parameters
    # raise ValueError
    with pytest.raises(ValueError):
        NLEIS_circuit = EISandNLEIS(
            circ_str_1, circ_str_2, initial_guess=initial_guess,
            constants={'L0': 1e-7, 'TDSn1_7': 0})
    # check that constants number beyond the range
    # of allowed number of parameters
    # raise ValueError
    with pytest.raises(ValueError):
        NLEIS_circuit = EISandNLEIS(
            circ_str_1, circ_str_2, initial_guess=initial_guess,
            constants={'L2': 1e-7, 'TDSn1_6': 0})

    # check incorrect element pair raise TypeError
    with pytest.raises(TypeError):
        NLEIS_circuit = EISandNLEIS(
            circ_str_1, circ_str_1, initial_guess=initial_guess,
            constants={'L0': 1e-7, 'TDSn1_6': 0})

    # check that constants either circ_str_1 or circ_str_2
    # cannot be empty/ raise ValueError if empty
    # check missing circuit_2 raise ValueError
    with pytest.raises(ValueError):
        NLEIS_circuit = EISandNLEIS(
            circ_str_1, circ_str_2='', initial_guess=initial_guess,
            constants={'L0': 1e-7, 'TDSn1_6': 0})

    # check that wrong length of initial_guess
    # with respect to the length of the circuit raise ValueError

    with pytest.raises(ValueError):
        NLEIS_circuit = EISandNLEIS(
            circ_str_1, circ_str_2, initial_guess=[1],
            constants={'L0': 1e-7, 'TDSn1_6': 0})

    # check that wrong constants input raise ValueError
    # when the wrong constants is applied for EIS_circuit

    with pytest.raises(ValueError):
        NLEIS_circuit = EISandNLEIS(
            circ_str_1, circ_str_2, initial_guess=initial_guess,
            constants={'L0': 1e-7, 'TDS1_10': 0})


def test_eq():
    NLEIS_circuit = NLEISCustomCircuit()
    simul_circuit = EISandNLEIS()

    with pytest.raises(TypeError):
        simul_circuit == NLEIS_circuit


def test_EISandNLEIS_fitting():
    circ_str_1 = 'L0-R0-TDS0-TDS1'
    circ_str_2 = 'd(TDSn0,TDSn1)'

    # Test example shown "Getting Started" page
    # Note: solution to the solid state diffuion might
    # converge to different value given the exact
    # envrionment used (i.e. large confidence interval)

    # The test framework here can be improved in the future
    # with better and stable initial guess
    results = [9.81368514e-08, 1.34551972e-02, 2.52387276e-02, 5.06176242e-03,
               8.82244297e+00, 8.70692162e-05, 3.55536976e+00, 1.22576118e+01,
               8.75169434e-02, 2.09045802e-02, 1.13804384e-03, 8.13658287e-01,
               1.83783329e+02, 3.20554700e+09, 1.02277512e+00, 6.39228801e-03]
    initial_guess = results

    # initial_guess
    # L0,RO
    # TDS0 + additioal nonlinear parameters
    # TDS1 + additioal nonlinear parameters

    NLEIS_circuit = EISandNLEIS(
        circ_str_1, circ_str_2, initial_guess=initial_guess)

    # test predict with initial_guess
    Z1_fit, Z2_fit = NLEIS_circuit.predict(f)
    assert np.allclose(Z1, Z1_fit, rtol=1e-2, atol=1e-2)
    assert np.allclose(Z2_trunc, Z2_fit, rtol=1e-2, atol=1e-2)

    # test fitting
    NLEIS_circuit.fit(f, Z1, Z2)
    p = NLEIS_circuit.parameters_

    assert np.allclose(p, results)

    # test the extract method
    dict1, dict2 = NLEIS_circuit.extract()
    assert list(dict1.keys()) == ['L0', 'R0',
                                  'TDS0_0', 'TDS0_1', 'TDS0_2', 'TDS0_3',
                                  'TDS0_4',
                                  'TDS1_0', 'TDS1_1', 'TDS1_2',
                                  'TDS1_3', 'TDS1_4']
    assert list(dict2.keys()) == ['TDSn0_0', 'TDSn0_1', 'TDSn0_2', 'TDSn0_3',
                                  'TDSn0_4', 'TDSn0_5',
                                  'TDSn0_6', 'TDSn1_0', 'TDSn1_1', 'TDSn1_2',
                                  'TDSn1_3', 'TDSn1_4', 'TDSn1_5', 'TDSn1_6']

    # test plotting
    # kind = {'nyquist', 'bode'} should return a plt.Axes() object
    _, ax = plt.subplots(1, 2)
    assert isinstance(NLEIS_circuit.plot(
        ax, None, Z1, Z2, kind='nyquist'), type(ax))
    assert isinstance(NLEIS_circuit.plot(
        None, f, Z1, Z2, kind='nyquist'), type(ax))
    _, axes = plt.subplots(2, 2)
    assert isinstance(NLEIS_circuit.plot(
        axes, f, Z1, Z2, kind='bode')[0], type(ax))
    assert isinstance(NLEIS_circuit.plot(
        None, f, Z1, Z2, kind='bode')[0], type(ax))

    # check altair plotting with a fit circuit
    chart1, chart2 = NLEIS_circuit.plot(f_data=f, Z1_data=Z1, Z2_data=Z2,
                                        kind='altair')

    datasets = json.loads(chart1.to_json())['datasets']
    for dataset in datasets.keys():
        assert len(datasets[dataset]) == len(Z1)

    datasets = json.loads(chart2.to_json())['datasets']
    for dataset in datasets.keys():
        assert len(datasets[dataset]) == len(Z2_trunc)

    # incorrect kind raises a ValueError
    with pytest.raises(ValueError):
        NLEIS_circuit.plot(None, f, Z1, Z2, kind='SomethingElse')

    # test optimization with negative log-likelihood
    initial_guess = [9.81387341e-08, 1.34551661e-02, 2.52404567e-02,
                     5.06142381e-03,
                     8.82333612e+00, 8.80981845e-05, 3.59757749e+00,
                     1.22590097e+01,
                     8.74790184e-02, 2.09035936e-02, 1.13816141e-03,
                     8.13599509e-01,
                     1.71140213e+02, 2.77968560e+09, 1.02284630e+00,
                     6.38829808e-03]
    NLEIS_circuit = EISandNLEIS(
        circ_str_1, circ_str_2, initial_guess=initial_guess)

    NLEIS_circuit.fit(f, Z1, Z2, opt='neg')
    p2 = NLEIS_circuit.parameters_
    results2 = [9.81387329e-08, 1.34551659e-02, 2.52509590e-02, 5.14440732e-03,
                8.82333612e+00, 8.80981835e-05, 3.59757749e+00, 1.22590097e+01,
                8.74790233e-02, 2.09083711e-02, 1.13829972e-03, 8.13599522e-01,
                1.71140213e+02, 2.77968560e+09, 1.02284629e+00, 6.38830393e-03]

    assert np.allclose(p2, results2, rtol=1e-3, atol=1e-3)


def test_NLEISCustomCircuit():
    circ_str = 'd(TDSn0,TDSn1)'
    initial_guess = [
        5e-3, 1e-3, 10, 1e-2, 100, 10, 0.1,
        # TDS0 + additioal nonlinear parameters
        1e-3, 1e-3, 1e-3, 1e-2, 1000, 0, 0,
        # TDS1 + additioal nonlinear parameters
    ]

    NLEIS_circuit = NLEISCustomCircuit(
        circ_str, initial_guess=initial_guess)

    assert not NLEIS_circuit._is_fit()

    # check get_param_names()
    full_names_NLEIS, all_units_NLEIS = NLEIS_circuit.get_param_names()
    assert full_names_NLEIS == ['TDSn0_0', 'TDSn0_1', 'TDSn0_2', 'TDSn0_3',
                                'TDSn0_4', 'TDSn0_5',
                                'TDSn0_6', 'TDSn1_0', 'TDSn1_1', 'TDSn1_2',
                                'TDSn1_3', 'TDSn1_4', 'TDSn1_5', 'TDSn1_6']
    assert all_units_NLEIS == ['Ohms', 'Ohms', 'F', 'Ohms', 's',
                               '1/V', '-', 'Ohms', 'Ohms', 'F', 'Ohms', 's',
                               '1/V', '-']

    # check _is_fit()
    assert not NLEIS_circuit._is_fit()

    # check complex frequencies raise TypeError
    with pytest.raises(TypeError):
        NLEIS_circuit.predict([0.42, 42 + 42j])

    # test is_fit method
    NLEIS_circuit.fit(f, Z2)
    assert NLEIS_circuit._is_fit()

    # test extract method
    dict = NLEIS_circuit.extract()
    assert list(dict.keys()) == ['TDSn0_0', 'TDSn0_1', 'TDSn0_2', 'TDSn0_3',
                                 'TDSn0_4', 'TDSn0_5',
                                 'TDSn0_6', 'TDSn1_0', 'TDSn1_1', 'TDSn1_2',
                                 'TDSn1_3', 'TDSn1_4', 'TDSn1_5', 'TDSn1_6']

    # test plotting
    # kind = {'nyquist', 'bode'} should return a plt.Axes() object
    _, ax = plt.subplots()
    assert isinstance(NLEIS_circuit.plot(
        ax, None, Z2, kind='nyquist'), type(ax))
    assert isinstance(NLEIS_circuit.plot(
        None, f, Z2, kind='nyquist'), type(ax))
    _, axes = plt.subplots(1, 2)
    assert isinstance(NLEIS_circuit.plot(
        axes, f, Z2, kind='bode')[0], type(ax))
    assert isinstance(NLEIS_circuit.plot(
        None, f, Z2, kind='bode')[0], type(ax))

    # check altair plotting with a fit circuit
    chart = NLEIS_circuit.plot(f_data=f, Z2_data=Z2, max_f=10,
                               kind='altair')

    datasets = json.loads(chart.to_json())['datasets']
    for dataset in datasets.keys():
        assert len(datasets[dataset]) == len(Z2_trunc)

    # incorrect kind raises a ValueError
    with pytest.raises(ValueError):
        NLEIS_circuit.plot(None, f, Z2, kind='SomethingElse')



import altair as alt
import matplotlib.pyplot as plt  # noqa: F401
import numpy as np
import pandas as pd
from matplotlib import ticker

# These are all slightly modified versin of the visualization code.
# Can be seamlessly inegrated into impedance.py in the future.


def plot_altair(data_dict, k=1, units='Ω', size=400, background='#FFFFFF'):
    """ Plots impedance as an interactive Nyquist/Bode plot using altair

        Parameters
        ----------
        data_dict: dict
            dictionary with keys
            'f': frequencies
            'Z': impedance
            'fmt': {'-' for a line with circles, else circles}
        k: int
            harmonics of the impedance data
        units: str
            units of the impedance data
        size: int
            size in pixels of Nyquist height/width
        background: str
            hex color string for chart background (default = '#FFFFFF')

        Returns
        -------
        chart: altair.Chart
    """

    Z_df = pd.DataFrame(columns=['f', 'z_real', 'z_imag', 'kind', 'fmt'])
    for kind in data_dict.keys():
        f = data_dict[kind]['f']
        Z = data_dict[kind]['Z']
        fmt = data_dict[kind].get('fmt', 'o')

        df = pd.DataFrame({'f': f, 'z_real': Z.real, 'z_imag': Z.imag,
                           'kind': kind, 'fmt': fmt})

        Z_df = pd.concat([Z_df, df], ignore_index=True)
        # Z_df.append(df)

    range_x = max(Z_df['z_real']) - min(Z_df['z_real'])
    range_y = max(-Z_df['z_imag']) - min(-Z_df['z_imag'])

    rng = max(range_x, range_y)

    min_x = min(Z_df['z_real'])
    max_x = min(Z_df['z_real']) + rng
    min_y = min(-Z_df['z_imag'])
    max_y = min(-Z_df['z_imag']) + rng

    nearest = alt.selection_single(on='mouseover', nearest=True,
                                   empty='none', fields=['f'],
                                   clear='mouseout')
    # potential future improvement
    # nearest = alt.selection_point(on='mouseover', nearest=True,
    #                               empty='none', fields=['f'])

    fmts = Z_df['fmt'].unique()
    nyquists, bode_mags, bode_phss = [], [], []
    if '-' in fmts:
        df = Z_df.groupby('fmt').get_group('-')
        # These are changed to introduce harmonics and units

        nyquist = alt.Chart(df).mark_circle().encode(
            x=alt.X('z_real:Q', axis=alt.Axis(
                title="Z{}' [{}]".format(k, units)),
                scale=alt.Scale(domain=[min_x, max_x],
                                nice=False, padding=5), sort=None),
            y=alt.Y('neg_z_imag:Q', axis=alt.Axis(
                title="-Z{}'' [{}]".format(k, units)),
                scale=alt.Scale(domain=[min_y, max_y],
                                nice=False, padding=5), sort=None),
            size=alt.condition(nearest, alt.value(80), alt.value(30)),

            color='kind:N',
            tooltip=[  # Added Tooltips to show values
                alt.Tooltip('kind:N', title="Kind"),
                alt.Tooltip('f:Q', title="f [Hz]"),
                alt.Tooltip('z_real:Q', title="Z{}' [{}]".format(k, units)),
                alt.Tooltip('z_imag:Q', title="Z{}'' [{}]".format(k, units))
            ]
        ).add_selection(
            nearest
        ).properties(
            height=size,
            width=size
        ).transform_calculate(
            neg_z_imag='-datum.z_imag'
        ).interactive()

        nyquist_line_plot = alt.Chart(df).mark_line().encode(
            x=alt.X('z_real:Q', axis=alt.Axis(
                title="Z{}' [{}]".format(k, units)),
                scale=alt.Scale(domain=[min_x, max_x],
                                nice=False, padding=5), sort=None),
            y=alt.Y('neg_z_imag:Q', axis=alt.Axis(
                title="-Z{}'' [{}]".format(k, units)),
                scale=alt.Scale(domain=[min_y, max_y],
                                nice=False, padding=5), sort=None),
            color='kind:N',
            order=alt.Order('f:Q', sort='ascending')
        ).properties(
            height=size,
            width=size
        ).transform_calculate(
            neg_z_imag='-datum.z_imag'
        )

        bode = alt.Chart(df).mark_circle().encode(
            alt.X('f:Q', axis=alt.Axis(title="f [Hz]"),
                  scale=alt.Scale(type='log', nice=False), sort=None),
            size=alt.condition(nearest, alt.value(80), alt.value(30)),

            color='kind:N',
        ).add_selection(
            nearest
        ).properties(
            width=size,
            height=size/2 - 25
        ).transform_calculate(
            mag="sqrt(pow(datum.z_real,2) + pow(datum.z_imag,2))",
            neg_phase="-(180/PI)*atan2(datum.z_imag,datum.z_real)"
        ).interactive()

        bode_mag = bode.encode(
            y=alt.Y('mag:Q', axis=alt.Axis(
                title="|Z{}| [{}]".format(k, units)), sort=None),
            tooltip=[  # Added Tooltips to show values
                alt.Tooltip('kind:N', title="Kind"),
                alt.Tooltip('f:Q', title="f [Hz]"),
                alt.Tooltip('mag:Q', title="|Z{}| [{}]".format(k, units))
            ]
        )
        bode_phs = bode.encode(
            y=alt.Y('neg_phase:Q', axis=alt.Axis(title="-ϕ{} [°]".format(k)),
                    sort=None),
            tooltip=[  # Added Tooltips to show values
                alt.Tooltip('kind:N', title="Kind"),
                alt.Tooltip('f:Q', title="f [Hz]"),
                alt.Tooltip('neg_phase:Q', title="-ϕ{} [°]".format(k)),
            ]
        )

        bode_line_plot = alt.Chart(df).mark_line().encode(
            alt.X('f:Q', axis=alt.Axis(title="f [Hz]"),
                  scale=alt.Scale(type='log', nice=False), sort=None),
            color='kind:N',
        ).properties(
            width=size,
            height=size/2 - 25
        ).transform_calculate(
            mag="sqrt(pow(datum.z_real,2) + pow(datum.z_imag,2))",
            neg_phase="-(180/PI)*atan2(datum.z_imag,datum.z_real)"
        )

        bode_mag_line_plot = bode_line_plot.encode(
            y=alt.Y('mag:Q', axis=alt.Axis(
                title="|Z{}| [{}]".format(k, units)), sort=None),
        )
        bode_phs_line_plot = bode_line_plot.encode(
            y=alt.Y('neg_phase:Q', axis=alt.Axis(title="-ϕ{} [°]".format(k)),
                    sort=None),
        )

        nyquists.append(nyquist)
        nyquists.append(nyquist_line_plot)
        bode_mags.append(bode_mag)
        bode_mags.append(bode_mag_line_plot)
        bode_phss.append(bode_phs)
        bode_phss.append(bode_phs_line_plot)

    if 'o' in fmts:
        df = Z_df.groupby('fmt').get_group('o')

        nyquist = alt.Chart(df).mark_circle().encode(
            x=alt.X('z_real:Q', axis=alt.Axis(
                title="Z{}' [{}]".format(k, units)),
                scale=alt.Scale(domain=[min_x, max_x],
                                nice=False, padding=5), sort=None),
            y=alt.Y('neg_z_imag:Q', axis=alt.Axis(
                title="-Z{}'' [{}]".format(k, units)),
                scale=alt.Scale(domain=[min_y, max_y],
                                nice=False, padding=5), sort=None),
            size=alt.condition(nearest, alt.value(80), alt.value(30)),
            color=alt.Color('kind:N', legend=alt.Legend(title='Legend')),
            tooltip=[  # Added Tooltips to show values
                alt.Tooltip('kind:N', title="Kind"),
                alt.Tooltip('f:Q', title="f [Hz]"),
                alt.Tooltip('z_real:Q', title="Z{}' [{}]".format(k, units)),
                alt.Tooltip('z_imag:Q', title="Z{}'' [{}]".format(k, units))
            ]
        ).add_selection(
            nearest
            # potential future improvement
            # ).add_params(
            #     nearest
        ).properties(
            height=size,
            width=size
        ).transform_calculate(
            neg_z_imag='-datum.z_imag'
        ).interactive()

        bode = alt.Chart(df).mark_circle().encode(
            alt.X('f:Q', axis=alt.Axis(title="f [Hz]"),
                  scale=alt.Scale(type='log', nice=False), sort=None),
            size=alt.condition(nearest, alt.value(80), alt.value(30)),
            color='kind:N'
        ).add_selection(
            nearest
            # potential future improvement
            # ).add_params(
            #     nearest
        ).properties(
            width=size,
            height=size/2 - 25
        ).transform_calculate(
            mag="sqrt(pow(datum.z_real,2) + pow(datum.z_imag,2))",
            neg_phase="-(180/PI)*atan2(datum.z_imag,datum.z_real)"
        ).interactive()

        bode_mag = bode.encode(
            y=alt.Y('mag:Q', axis=alt.Axis(
                title="|Z{}| [{}]".format(k, units)), sort=None),
            tooltip=[  # Added Tooltips to show values
                alt.Tooltip('kind:N', title="Kind"),
                alt.Tooltip('f:Q', title="f [Hz]"),
                alt.Tooltip('mag:Q', title="|Z{}| [{}]".format(k, units))
            ])
        bode_phs = bode.encode(
            y=alt.Y('neg_phase:Q', axis=alt.Axis(title="-ϕ{} [°]".format(k)),
                    sort=None),
            tooltip=[  # Added Tooltips to show values
                alt.Tooltip('kind:N', title="Kind"),
                alt.Tooltip('f:Q', title="f [Hz]"),
                alt.Tooltip('neg_phase:Q',
                            title="-ϕ{} [°]".format(k)),
            ])

        nyquists.append(nyquist)
        bode_mags.append(bode_mag)
        bode_phss.append(bode_phs)

    full_bode = alt.layer(*bode_mags) & alt.layer(*bode_phss)

    return (full_bode | alt.layer(*nyquists)).configure(background=background)


def plot_first(ax, Z, scale=1, fmt='.-', labelsize=20,
               ticksize=14, **kwargs):
    """
    Plots EIS impedance as a Nyquist plot using matplotlib.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Axes on which to plot the Nyquist plot.

    Z : numpy.ndarray of complex
        EIS data to be plotted. The real part is plotted on the x-axis,
        and the negative imaginary part on the y-axis.

    scale : float, optional
        Scaling factor for the axes. The default is 1.

    fmt : str, optional
        Format string passed to matplotlib (e.g., '.-' for line style, 'o' for
        markers). The default is '.-'.

    labelsize : int, optional
        Font size for axis labels. The default is 20.

    ticksize : int, optional
        Font size for axis tick labels. The default is 14.

    Other Parameters
    ----------------
    **kwargs : `matplotlib.pyplot.Line2D` properties, optional
        Used to specify line properties like linewidth, line color, marker
        color, and line labels.

    Returns
    -------
    ax : matplotlib.axes.Axes
        The axes with the Nyquist plot for EIS.
    """

    ax.plot(np.real(Z), -np.imag(Z), fmt, **kwargs)

    # Make the axes square
    ax.set_aspect(aspect=1, anchor='C', adjustable='datalim')

    # Set the labels to -imaginary vs real
    ax.set_xlabel(r'$\tilde{Z}_{1}^{\prime}(\omega)$' +
                  r' [$\Omega$]', fontsize=labelsize)
    ax.set_ylabel(r'$-\tilde{Z}_{1}^{\prime\prime}(\omega)$' +
                  r' [$\Omega$]', fontsize=labelsize)

    # Make the tick labels larger
    ax.tick_params(axis='both', which='major', labelsize=ticksize)

    # Change the number of labels on each axis to five
    ax.locator_params(axis='x', nbins=5, tight=True)
    ax.locator_params(axis='y', nbins=5, tight=True)

    # Add a light grid
    ax.grid(visible=True, which='major', axis='both', alpha=.5)

    # Change axis units to 10**log10(scale) and resize the offset text
    # ax.ticklabel_format(style='sci', axis='both')

    formatter = ticker.ScalarFormatter(useMathText=True)
    formatter.set_scientific(True)
    formatter.set_powerlimits((-1, 1))
    ax.yaxis.set_major_formatter(formatter)
    ax.xaxis.set_major_formatter(formatter)
    limits = -np.log10(scale)
    if limits != 0:
        ax.ticklabel_format(style='sci', axis='both',
                            scilimits=(limits, limits))
    y_offset = ax.yaxis.get_offset_text()
    y_offset.set_size(18)
    t = ax.xaxis.get_offset_text()
    t.set_size(18)

    return ax


def plot_second(ax, Z, scale=1, fmt='.-', labelsize=20,
                ticksize=14, **kwargs):
    """
    Plots 2nd-NLEIS impedance as a Nyquist plot using matplotlib.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Axes on which to plot the Nyquist plot.

    Z : numpy.ndarray of complex
        2nd-NLEIS data to be plotted. The real part is plotted on the x-axis,
        and the negative imaginary part on the y-axis.

    scale : float, optional
        Scaling factor for the axes. The default is 1.

    fmt : str, optional
        Format string passed to matplotlib (e.g., '.-' for line style, 'o' for
        markers). The default is '.-'.

    labelsize : int, optional
        Font size for axis labels. The default is 20.

    ticksize : int, optional
        Font size for axis tick labels. The default is 14.

    Other Parameters
    ----------------
    **kwargs : `matplotlib.pyplot.Line2D` properties, optional
        Used to specify line properties like linewidth, line color, marker
        color, and line labels.

    Returns
    -------
    ax : matplotlib.axes.Axes
        The axes with the Nyquist plot for 2nd-NLEIS.
    """

    ax.plot(np.real(Z), -np.imag(Z), fmt, **kwargs)

    # Make the axes square
    ax.set_aspect(aspect=1, anchor='C', adjustable='datalim')

    # Set the labels to -imaginary vs real
    ax.set_xlabel(r'$\tilde{Z}_{2}^{\prime}(\omega)$' +
                  r' [$\Omega / A$]', fontsize=labelsize)
    ax.set_ylabel(r'$-\tilde{Z}_{2}^{\prime\prime}(\omega)$' +
                  r' [$\Omega / A$]', fontsize=labelsize)
    # Make the tick labels larger
    ax.tick_params(axis='both', which='major', labelsize=ticksize)

    # Change the number of labels on each axis to five
    ax.locator_params(axis='x', nbins=5, tight=True)
    ax.locator_params(axis='y', nbins=5, tight=True)

    # Add a light grid
    ax.grid(visible=True, which='major', axis='both', alpha=.5)

    # Change axis units to 10**log10(scale) and resize the offset text

    formatter = ticker.ScalarFormatter(useMathText=True)
    formatter.set_scientific(True)
    formatter.set_powerlimits((-1, 1))
    ax.yaxis.set_major_formatter(formatter)
    ax.xaxis.set_major_formatter(formatter)

    limits = -np.log10(scale)
    if limits != 0:
        ax.ticklabel_format(style='scientific', axis='both',
                            scilimits=(limits, limits))
    y_offset = ax.yaxis.get_offset_text()
    y_offset.set_size(18)
    t = ax.xaxis.get_offset_text()
    t.set_size(18)

    return ax



from .nleis_fitting import simul_fit, \
    wrappedImpedance, individual_parameters
from .nleis_elements_pair import *  # noqa: F401, F403
from impedance.models.circuits.fitting import check_and_eval
from impedance.models.circuits.elements import circuit_elements, \
    get_element_from_name
from impedance.models.circuits.circuits import BaseCircuit
from .fitting import circuit_fit, buildCircuit, calculateCircuitLength, \
    extract_circuit_elements

from impedance.visualization import plot_bode
from .visualization import plot_altair, plot_first, plot_second
from nleis.fitting import CircuitGraph


import json
import matplotlib.pyplot as plt
import numpy as np
import warnings


class EISandNLEIS:
    # ToDO: add SSO method and SO method
    def __init__(self, circuit_1='', circuit_2='', initial_guess=[],
                 constants=None, name=None, graph=False, **kwargs):
        """
            Constructor for a customizable linear and nonlinear
            equivalent circuit model
            to enable simultaneous EIS and 2nd-NLEIS analysis.

            Parameters
            ----------
            circuit_1 : str
                A string representing the equivalent circuit for linear EIS.

            circuit_2 : str
                A string representing the equivalent circuit for 2nd-NLEIS.

            initial_guess : numpy.ndarray
                Initial guess values for the circuit parameters.

            constants : dict, optional
                Dictionary of parameters and values
                to hold constant during fitting (e.g., {"R0": 0.1}).
                These constants can apply to both EIS and 2nd-NLEIS elements.
                Note:
                The initial guess for constant parameters should be excluded.

            name : str, optional
                A name for the model.
            graph : bool, optional
                Whether to use execution graph to process the circuit.
                Defaults to False, which uses eval based code

            Notes
            -----
            The custom circuit for EIS is defined as a string with elements
            in series (separated by `-`) and
            elements in parallel (grouped as p(x, y)).

            For 2nd-NLEIS, the circuit should be grouped by `d(cathode, anode)`
            to denote difference calculations.

            Each element can be appended with an integer (e.g., R0)
            or an underscore and an integer (e.g., CPE_1)
            to help track multiple elements of the same type.

            Example
            -------
            A two-electrode cell with a spherical porous cathode and anode,
            resistor, and inductor can be represented as:

            EIS: circuit_1 = 'L0-R0-TDS0-TDS1'
            NLEIS: circuit_2 = 'd(TDSn0-TDSn1)'

        """
        self.graph = graph
        # if supplied, check that initial_guess is valid and store
        initial_guess = [x for x in initial_guess if x is not None]
        for i in initial_guess:
            if not isinstance(i, (float, int, np.int32, np.float64)):
                raise TypeError(f'value {i} in initial_guess is not a number')

        # initalize class attributes
        # self.initial_guess = list(initial_guess)
        self.initial_guess = initial_guess

        self.circuit_1 = circuit_1
        self.circuit_2 = circuit_2
        elements_1 = extract_circuit_elements(circuit_1)
        elements_2 = extract_circuit_elements(circuit_2)

        # new code for circuit length calculation using circuit element
        # producing edited circuit
        # i.e. circuit_1 = L0-R0-TDP0-TDS1;
        # circuit_2 = TDPn0-TDSn1; edited_circuit = L0-R0-TDPn0-TDSn1
        if elements_1 != [''] or elements_2 != ['']:
            if elements_1 == [''] or elements_2 == ['']:
                raise ValueError(
                    'Either circuit_1 or circuit_2 cannot be empty')
            edited_circuit = ''
            for elem in elements_1:
                raw_elem = get_element_from_name(elem)
                len_raw_elem = len(raw_elem)
                nl_elem = elem[0:len_raw_elem] + 'n' + elem[len_raw_elem:]
                # elem[0:-1]+'n'+elem[-1]
                if nl_elem in elements_2:
                    edited_circuit += '-' + nl_elem
                else:
                    edited_circuit += '-' + elem
            self.edited_circuit = edited_circuit[1:]
        else:
            self.edited_circuit = ''

        circuit_len = calculateCircuitLength(self.edited_circuit)

        for element in elements_2:
            if ((element.replace('n', '') not in elements_1) or
                    ('n' not in element and element != '')):
                raise TypeError(
                    'The pairing of linear and nonlinear elements must be'
                    + 'presented correctly for simultaneous fitting.'
                    + ' Double check the EIS circuit: '
                    + f'{circuit_1}' + ' and the NLEIS circuit: '
                    + f'{circuit_2}' + '. The parsed elements are '
                    + f'{elements_1}'+' for circuit_1, and '
                    + f'{elements_2}' + ' for circuit_2.')

        input_elements = elements_1 + elements_2

        if constants is not None:
            self.constants = constants
            self.constants_1 = {}
            self.constants_2 = {}
            # New code for constant separatation
            # in the current development the differentiation
            # between linear and nonlinear is separated by 'n'
            # using get_element_from_name to get the raw element
            # and adding n to the end

            # The constant for EIS circuit will present in
            # both constants_1 and constants_2
            # The constants for 2nd-NLEIS will only present in
            # constants_2
            for elem in self.constants:

                raw_elem = get_element_from_name(elem)
                raw_circuit_elem = elem.split('_')[0]

                if raw_circuit_elem not in input_elements:
                    raise ValueError(f'{raw_elem} not in ' +
                                     f'input elements ({input_elements})')
                raw_num_params = check_and_eval(raw_elem).num_params
                if raw_num_params <= 1:
                    # currently there is no single parameter nonlinear element,
                    # so this can work, but might be changed in the future
                    self.constants_1[elem] = self.constants[elem]
                else:
                    param_num = int(elem.split('_')[-1])
                    if raw_elem[-1] != 'n':

                        if param_num >= raw_num_params:
                            raise ValueError(
                                f'{elem} is out of the range of'
                                + ' the maximum allowed '
                                + f'number of parameters ({raw_num_params})')

                        self.constants_1[elem] = self.constants[elem]
                        len_elem = len(raw_elem)
                        nl_elem = elem[0:len_elem]+'n'+elem[len_elem:]
                        raw_nl_elem = get_element_from_name(nl_elem)
                        allowed_elems = circuit_elements.keys()
                        if raw_nl_elem in allowed_elems:
                            self.constants_2[nl_elem] = self.constants[elem]
                        else:
                            # this code is kept here to ignore
                            # EIS only constants in constnats_2
                            self.constants_2[elem] = self.constants[elem]

                    if raw_elem[-1] == 'n':

                        if param_num >= raw_num_params:
                            raise ValueError(
                                f'{elem} is out of the range of'
                                + ' the maximum allowed '
                                + f'number of parameters ({raw_num_params})')

                        num_params = check_and_eval(raw_elem[0:-1]).num_params
                        if param_num < num_params:
                            self.constants_1[elem.replace(
                                'n', '')] = self.constants[elem]
                        self.constants_2[elem] = self.constants[elem]

        else:
            self.constants = {}
            self.constants_1 = {}
            self.constants_2 = {}

        if len(self.initial_guess) + len(self.constants) != circuit_len:
            raise ValueError('The number of initial guesses ' +
                             f'({len(self.initial_guess)}) + ' +
                             'the number of constants ' +
                             f'({len(self.constants)})' +
                             ' must be equal to ' +
                             f'the circuit length ({circuit_len})')
        self.name = name

        # initialize fit parameters and confidence intervals
        self.parameters_ = None
        self.conf_ = None

        self.p1, self.p2 = individual_parameters(
            self.edited_circuit, self.initial_guess,
            self.constants_1, self.constants_2)
        if self.circuit_1:
            self.cg1 = CircuitGraph(self.circuit_1, self.constants_1)
        if self.circuit_2:
            self.cg2 = CircuitGraph(self.circuit_2, self.constants_2)

    def __eq__(self, other):
        if self.__class__ == other.__class__:
            matches = []
            for key, value in self.__dict__.items():
                if isinstance(value, np.ndarray):
                    matches.append((value == other.__dict__[key]).all())
                else:
                    matches.append(value == other.__dict__[key])
            return np.array(matches).all()
        else:
            raise TypeError('Comparing object is not of the same type.')

    def fit(self, frequencies, Z1, Z2, bounds=None,
            opt='max', max_f=10, cost=0.5, param_norm=True,
            positive=True, **kwargs):
        """

        Fit the EIS and 2nd-NLEIS circuit model simultaneously.

        Parameters
        ----------
        frequencies : numpy.ndarray
            Array of frequencies.

        Z1 : numpy.ndarray, dtype=complex128
            EIS values to fit.

        Z2 : numpy.ndarray, dtype=complex128
            2nd-NLEIS values to fit.

        bounds : tuple of array_like, optional
            A 2-tuple representing the lower and upper bounds on parameters.
            If bounds are provided, the input will be normalized to stabilize
            the algorithm.

        opt : str, optional
            Optimization method to use. Default is 'max'. 'max' refers to the
            maximum normalization method introduced by Ji and Schwartz. [1]

            The 'neg' option is available to perform negative log-likelihood
            calculation, as introduced by Kirk et al. [2]

        max_f : float, optional
            The maximum frequency of interest for 2nd-NLEIS. Default is 10.

        cost : float, optional
            Cost function weight for optimization. Default is 0.5.
            Overall cost = cost * EIS + (1 - cost) * 2nd-NLEIS.

        param_norm : bool, optional
            Whether to apply parameter normalization. Default is True.

        positive : bool, optional
            Whether to restrict the EIS values to the positive real quadrant.
            Default is True.

        kwargs :
            Additional keyword arguments passed to `simul_fit`,
            and subsequently to `scipy.optimize.curve_fit`.

        Returns
        -------
        self : object
            Returns the instance of the model for chaining.

        References
        ----------
        [1] Y. Ji, D.T. Schwartz,
        Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy:
        II.Model-Based Analysis of Lithium-Ion Battery Experiments.
        J. Electrochem. Soc., 2024. `doi:10.1149/1945-7111/ad2596
        <https://iopscience.iop.org/article/10.1149/1945-7111/ad2596>`_.

        [2] Kirk et al.,
        Nonlinear Electrochemical Impedance Spectroscopy for Lithium-Ion
        Battery Model Parameterization.
        J. Electrochem. Soc., 2023. `doi:10.1149/1945-7111/acada7
        <https://iopscience.iop.org/article/10.1149/1945-7111/acada7>`_.

        """
        # check that inputs are valid:
        #    frequencies: array of numbers
        #    impedance: array of complex numbers
        #    impedance and frequency match in length

        frequencies = np.array(frequencies, dtype=float)
        Z1 = np.array(Z1, dtype=complex)
        Z2 = np.array(Z2, dtype=complex)
        if len(frequencies) != len(Z1):
            raise ValueError(
                'length of frequencies and impedance do not match for EIS')

        if self.initial_guess != []:
            parameters, conf = simul_fit(
                frequencies, Z1, Z2, self.circuit_1, self.circuit_2,
                self.edited_circuit, self.initial_guess,
                constants_1=self.constants_1,
                constants_2=self.constants_2, bounds=bounds, opt=opt,
                cost=cost, max_f=max_f, param_norm=param_norm,
                positive=positive, graph=self.graph,
                **kwargs)
            # self.parameters_ = list(parameters)
            self.parameters_ = parameters

            if conf is not None:
                # self.conf_ = list(conf)
                self.conf_ = conf
                self.conf1, self.conf2 = individual_parameters(
                    self.edited_circuit, self.conf_, self.constants_1,
                    self.constants_2)
            # if cov is not None:
            #     self.cov_ = cov
            #     self.cov1, self.cov2 = individual_parameters(
            #         self.edited_circuit, self.cov_, self.constants_1,
            #         self.constants_2)

            self.p1, self.p2 = individual_parameters(
                self.edited_circuit, self.parameters_, self.constants_1,
                self.constants_2)
        else:
            raise ValueError('no initial guess supplied')

        return self

    def _is_fit(self):
        """

        check if model has been fit (parameters_ is not None)

        """
        if self.parameters_ is not None:
            return True
        else:
            return False

    def predict(self, frequencies, max_f=10, use_initial=False):
        """
        Predict EIS and 2nd-NLEIS using
        linear and nonlinear equivalent circuit model.

        Parameters
        ----------
        frequencies : numpy.ndarray
            Array of frequency values.

        max_f : float, optional
            The maximum frequency of interest for 2nd-NLEIS. Default is 10.

        use_initial : bool
            If True and the model was previously fit,
            use the initial parameters instead of the fitted ones.

        Returns
        -------
        x1 : numpy.ndarray
            Calculated impedance for EIS (Z1).

        x2 : numpy.ndarray
            Calculated impedance for 2nd-NLEIS (Z2).

        """

        if not isinstance(frequencies, np.ndarray):
            raise TypeError('frequencies is not of type np.ndarray')
        if not (np.issubdtype(frequencies.dtype, np.integer) or
                np.issubdtype(frequencies.dtype, np.floating)):
            raise TypeError('frequencies array should have a numeric ' +
                            f'dtype (currently {frequencies.dtype})')

        f1 = frequencies
        mask = np.array(frequencies) < max_f
        f2 = frequencies[mask]

        if self._is_fit() and not use_initial:
            x1, x2 = wrappedImpedance(self.edited_circuit,
                                      self.circuit_1, self.constants_1,
                                      self.circuit_2, self.constants_2, f1, f2,
                                      self.parameters_, graph=self.graph)
            return x1, x2
        else:
            warnings.warn("Simulating circuit based on initial parameters")
            x1, x2 = wrappedImpedance(self.edited_circuit,
                                      self.circuit_1, self.constants_1,
                                      self.circuit_2, self.constants_2, f1, f2,
                                      self.initial_guess, graph=self.graph)
            return x1, x2

    def get_param_names(self, circuit, constants):
        """
        Converts a circuit string into
        a list of parameter names and their units.

        Parameters
        ----------
        circuit : str
            A string representing the circuit. This string may contain elements
            connected in series or parallel, and may also include
            nonlinear terms represented by 'd' for NLEIS elements.

        constants : dict
            A dictionary of constant parameters with their values.
            Parameters that are constants are excluded from the output.

        Returns
        -------
        full_names : list of str
            A list containing the full names of the parameters derived from the
            circuit string.

        all_units : list of str
            A list containing the corresponding units of the parameters.
        """

        # parse the element names from the circuit string
        names = circuit.replace('d', '').replace(
            '(', '').replace(')', '')  # edit for nleis.py

        names = names.replace('p', '').replace('(', '').replace(')', '')
        names = names.replace(',', '-').replace(' ', '').split('-')

        full_names, all_units = [], []
        for name in names:
            elem = get_element_from_name(name)
            num_params = check_and_eval(elem).num_params
            units = check_and_eval(elem).units
            if num_params > 1:
                for j in range(num_params):
                    full_name = '{}_{}'.format(name, j)
                    if full_name not in constants.keys():
                        full_names.append(full_name)
                        all_units.append(units[j])
            else:
                if name not in constants.keys():
                    full_names.append(name)
                    all_units.append(units[0])

        return full_names, all_units

    def __str__(self):
        """

        Defines the pretty printing of the circuit
        for both EIS and 2nd-NLEIS

        """

        to_print = '\n'
        if self.name is not None:
            to_print += 'Name: {}\n'.format(self.name)
        to_print += 'EIS Circuit string: {}\n'.format(self.circuit_1)
        to_print += 'NLEIS Circuit string: {}\n'.format(self.circuit_2)
        to_print += "Fit: {}\n".format(self._is_fit())

        if len(self.constants) > 0:
            to_print += '\nConstants:\n'
            for name, value in self.constants.items():
                elem = get_element_from_name(name)
                units = check_and_eval(elem).units
                if '_' in name:
                    unit = units[int(name.split('_')[-1])]
                else:
                    unit = units[0]
                to_print += '  {:>5} = {:.2e} [{}]\n'.format(name, value, unit)

        names1, units1 = self.get_param_names(self.circuit_1, self.constants_1)
        names2, units2 = self.get_param_names(self.circuit_2, self.constants_2)

        to_print += '\nEIS Initial guesses:\n'
        p1, p2 = individual_parameters(
            self.edited_circuit, self.initial_guess, self.constants_1,
            self.constants_2)
        for name, unit, param in zip(names1, units1, p1):
            to_print += '  {:>5} = {:.2e} [{}]\n'.format(name, param, unit)
        to_print += '\nNLEIS Initial guesses:\n'
        for name, unit, param in zip(names2, units2, p2):
            to_print += '  {:>5} = {:.2e} [{}]\n'.format(name, param, unit)
        if self._is_fit():
            params1, confs1 = self.p1, self.conf1
            to_print += '\nEIS Fit parameters:\n'
            for name, unit, param, conf in zip(names1, units1,
                                               params1, confs1):
                to_print += '  {:>5} = {:.2e}'.format(name, param)
                to_print += '  (+/- {:.2e}) [{}]\n'.format(conf, unit)
            params2, confs2 = self.p2, self.conf2
            to_print += '\nNLEIS Fit parameters:\n'
            for name, unit, param, conf in zip(names2, units2,
                                               params2, confs2):
                to_print += '  {:>5} = {:.2e}'.format(name, param)
                to_print += '  (+/- {:.2e}) [{}]\n'.format(conf, unit)

        return to_print

    def extract(self):
        """
        Extracts the parameters of
        EIS and 2nd-NLEIS circuits into dictionaries.

        This method retrieves the parameter names from
        `circuit_1` (EIS circuit) and `circuit_2` (NLEIS circuit),
        and maps the fitted parameters to their respective names.

        Returns
        -------
        dict1 : dict
            A dictionary containing the parameters and
            their values for `circuit_1` (EIS circuit).

        dict2 : dict
            A dictionary containing the parameters and
            their values for `circuit_2` (NLEIS circuit).
        """

        names1, units1 = self.get_param_names(self.circuit_1, self.constants_1)
        dict1 = {}
        if self._is_fit():
            params1 = self.p1
            for names, param in zip(names1, params1):
                dict1[names] = param

        names2, units2 = self.get_param_names(self.circuit_2, self.constants_2)
        dict2 = {}
        if self._is_fit():
            params2 = self.p2

            for names, param in zip(names2, params2):
                dict2[names] = param

        return dict1, dict2

    def plot(self, ax=None, f_data=None, Z1_data=None, Z2_data=None,
             kind='nyquist', max_f=10, **kwargs):
        """
        Visualizes the model and optional data as Nyquist, Bode,
        or Altair (interactive) plots.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            Axes to plot on. If None, a new figure and axes will be created.

        f_data : numpy.ndarray, optional
            Array of frequency values for input data (used for Bode plots).
            The default is None.

        Z1_data : numpy.ndarray of complex, optional
            Array of EIS values (impedance data). The default is None.

        Z2_data : numpy.ndarray of complex, optional
            Array of 2nd-NLEIS values (impedance data). The default is None.

        kind : {'altair', 'nyquist', 'bode'}, optional
            The type of plot to visualize.

            - 'nyquist': Nyquist plot of real vs imaginary impedance.

            - 'bode': Bode plot showing magnitude and phase.

            - 'altair': Altair plot for interactive visualizations.

            Default is 'nyquist'.

        max_f : float, optional
            The maximum frequency of interest for 2nd-NLEIS data.
            The default is 10.

        **kwargs : optional
            Additional keyword arguments passed to `matplotlib.pyplot.Line2D`
            (for 'nyquist' or 'bode') to specify properties like linewidth,
            color, marker type, etc.
            If `kind` is 'altair',
            `kwargs` is used to specify plot height as `size`.

        Returns
        -------
        ax : matplotlib.axes.Axes or tuple of Axes, optional
            Axes object(s) with the plotted data
            if 'nyquist' or 'bode' plot is used.

        chart1, chart2 : altair.Chart, optional
            If `kind` is 'altair', it returns two Altair chart objects,
            one for EIS data and one for 2nd-NLEIS data.

        Raises
        ------
        ValueError
            If an unsupported `kind` is provided.
        """

        if kind == 'nyquist':
            if ax is None:
                _, ax = plt.subplots(1, 2, figsize=(10, 5))

            # we don't need the if else statement if we want
            # to enable plot without fit
            # if self._is_fit():
            if f_data is not None:
                f_pred = f_data
            else:
                f_pred = np.logspace(5, -3)

            if Z1_data is not None:
                ax[0] = plot_first(ax[0], Z1_data, scale=1, fmt='s', **kwargs)
                # impedance.py style
                # plot_nyquist(Z1_data, ls='', marker='s', ax=ax[0], **kwargs)
            if Z2_data is not None:
                if f_data is not None:
                    mask = np.array(f_data) < max_f
                    ax[1] = plot_second(ax[1], Z2_data[mask],
                                        scale=1, fmt='s', **kwargs)
                else:
                    ax[1] = plot_second(ax[1], Z2_data,
                                        scale=1, fmt='s', **kwargs)
                # impedance.py style
                # plot_nyquist(Z2_data, units='Ohms/A', ls='',
                # marker='s', ax=ax[1], **kwargs)

            Z1_fit, Z2_fit = self.predict(f_pred, max_f=max_f)
            ax[0] = plot_first(ax[0], Z1_fit, scale=1, fmt='-', **kwargs)
            # plot_nyquist(Z1_fit, ls='-', marker='', ax=ax[0], **kwargs)
            ax[1] = plot_second(ax[1], Z2_fit, scale=1, fmt='-', **kwargs)
            # plot_nyquist(Z2_fit,units='Ohms/A', ls='-',
            # marker='', ax=ax[1], **kwargs)

            ax[0].legend(['Data', 'Fit'])
            ax[1].legend(['Data', 'Fit'])
            return ax
        elif kind == 'bode':
            if ax is None:
                _, ax = plt.subplots(2, 2, figsize=(8, 8))

            if f_data is not None:
                f_pred = f_data
            else:
                f_pred = np.logspace(5, -3)

            if Z1_data is not None:
                if f_data is None:
                    raise ValueError('f_data must be specified if' +
                                     ' Z_data for a Bode plot')
                ax[:, 0] = plot_bode(f_data, Z1_data, ls='', marker='s',
                                     axes=ax[:, 0], **kwargs)
            if Z2_data is not None:
                if f_data is None:
                    raise ValueError('f_data must be specified if' +
                                     ' Z_data for a Bode plot')
                mask = np.array(f_pred) < max_f
                f2 = f_data[mask]
                Z2 = Z2_data[mask]
                ax[:, 1] = plot_bode(f2, Z2, units='Ω/A', ls='', marker='s',
                                     axes=ax[:, 1], **kwargs)
            # we don't need the if else statement
            # if we want to enable plot without fit
            # if self._is_fit():
            Z1_fit, Z2_fit = self.predict(f_pred, max_f=max_f)

            f1 = f_pred
            f2 = f_pred[np.array(f_pred) < max_f]

            ax[:, 0] = plot_bode(f1, Z1_fit, ls='-', marker='o',
                                 axes=ax[:, 0], **kwargs)
            ax[:, 1] = plot_bode(f2, Z2_fit, units='Ω/A', ls='-',
                                 marker='o',
                                 axes=ax[:, 1], **kwargs)

            ax[0, 0].set_ylabel(r'$|Z_{1}(\omega)|$ ' +
                                '$[{}]$'.format('Ω'), fontsize=20)
            ax[1, 0].set_ylabel(
                r'$-\phi_{Z_{1}}(\omega)$ ' + r'$[^o]$', fontsize=20)
            ax[0, 1].set_ylabel(r'$|Z_{2}(\omega)|$ ' +
                                '$[{}]$'.format('Ω/A'), fontsize=20)
            ax[1, 1].set_ylabel(
                r'$-\phi_{Z_{2}}(\omega)$ ' + r'$[^o]$', fontsize=20)
            ax[0, 0].legend(['Data', 'Fit'], fontsize=20)
            ax[0, 1].legend(['Data', 'Fit'], fontsize=20)
            ax[1, 0].legend(['Data', 'Fit'], fontsize=20)
            ax[1, 1].legend(['Data', 'Fit'], fontsize=20)
            return ax
        elif kind == 'altair':
            plot_dict_1 = {}
            plot_dict_2 = {}

            if ((Z1_data is not None) and (Z2_data is not None) and (
                    f_data is not None)):
                plot_dict_1['data'] = {'f': f_data, 'Z': Z1_data}
                mask = np.array(f_data) < max_f
                plot_dict_2['data'] = {'f': f_data[mask], 'Z': Z2_data[mask]}
            # we don't need the if else statement
            # if we want to enable plot without fit
            # if self._is_fit():
            if f_data is not None:

                f_pred = f_data

            else:

                f_pred = np.logspace(5, -3)

            if self.name is not None:
                name = self.name
            else:
                name = 'fit'

            Z1_fit, Z2_fit = self.predict(f_pred, max_f=max_f)
            mask = np.array(f_pred) < max_f
            plot_dict_1[name] = {'f': f_pred, 'Z': Z1_fit, 'fmt': '-'}
            plot_dict_2[name] = {'f': f_pred[mask], 'Z': Z2_fit, 'fmt': '-'}

            chart1 = plot_altair(plot_dict_1, k=1, units='Ω', **kwargs)
            chart2 = plot_altair(plot_dict_2, k=2, units='Ω/A', **kwargs)

            return chart1, chart2
        else:
            raise ValueError("Kind must be one of 'altair'," +
                             f"'nyquist', or 'bode' (received {kind})")

    def save(self, filepath):
        """
        Exports the model to a JSON file.

        This method saves the current model configuration,
        including circuit strings,
        initial guesses, constants, and (if fitted) the model parameters
        and their confidence intervals, to a JSON file.

        Parameters
        ----------
        filepath : str
            The file path where the model should be saved. The model will be
            serialized to a JSON file at this location.

        """

        model_string_1 = self.circuit_1
        model_string_2 = self.circuit_2
        edited_circuit_str = self.edited_circuit
        model_name = self.name

        initial_guess = self.initial_guess

        if self._is_fit():
            parameters_ = list(self.parameters_)
            model_conf_ = list(self.conf_)
            # model_cov_ = list(self.cov_)
            # parameters_ = self.parameters_
            # model_conf_ = self.conf_
            data_dict = {"Name": model_name,
                         "Circuit String 1": model_string_1,
                         "Circuit String 2": model_string_2,
                         "Initial Guess": initial_guess,
                         "Constants": self.constants,
                         "Constants 1": self.constants_1,
                         "Constants 2": self.constants_2,
                         "Fit": True,
                         "Parameters": parameters_,
                         "Confidence": model_conf_,
                         #  "Covariance": model_cov_,
                         "Edited Circuit Str": edited_circuit_str
                         }
        else:
            data_dict = {"Name": model_name,
                         "Circuit String 1": model_string_1,
                         "Circuit String 2": model_string_2,
                         "Initial Guess": initial_guess,
                         "Constants": self.constants,
                         "Constants 1": self.constants_1,
                         "Constants 2": self.constants_2,
                         "Edited Circuit Str": edited_circuit_str,
                         "Fit": False}

        with open(filepath, 'w') as f:
            json.dump(data_dict, f)

    def load(self, filepath, fitted_as_initial=False):
        """
        Imports a model from a JSON file.

        This method loads a saved model from a JSON file
        and restores its circuit strings, initial guesses,
        constants, and (if fitted) parameters. The user
        can choose whether to load the fitted parameters as the initial guess.

        Parameters
        ----------
        filepath : str
            The file path to the JSON file from which to load the model.

        fitted_as_initial : bool, optional
            If True, the model's fitted parameters will be
            loaded as initial guesses.
            Otherwise, the model will load both the initial
            and fitted parameters as a completed model. Default is False.

        """

        json_data_file = open(filepath, 'r')
        json_data = json.load(json_data_file)

        model_name = json_data["Name"]
        model_string_1 = json_data["Circuit String 1"]
        model_string_2 = json_data["Circuit String 2"]
        model_initial_guess = json_data["Initial Guess"]
        model_constants = json_data["Constants"]

        self.initial_guess = model_initial_guess
        self.circuit_1 = model_string_1
        self.circuit_2 = model_string_2
        self.edited_circuit = json_data["Edited Circuit Str"]

        self.constants = model_constants
        self.constants_1 = json_data["Constants 1"]
        self.constants_2 = json_data["Constants 2"]
        self.p1, self.p2 = individual_parameters(
            self.edited_circuit, self.initial_guess,
            self.constants_1, self.constants_2)

        self.cg1 = CircuitGraph(self.circuit_1, self.constants_1)
        self.cg2 = CircuitGraph(self.circuit_2, self.constants_2)

        self.name = model_name

        if json_data["Fit"]:
            if fitted_as_initial:
                self.initial_guess = json_data['Parameters']
                self.p1, self.p2 = individual_parameters(
                    self.edited_circuit, self.initial_guess,
                    self.constants_1, self.constants_2)

            else:
                self.parameters_ = np.array(json_data["Parameters"])
                self.conf_ = np.array(json_data["Confidence"])
                self.conf1, self.conf2 = individual_parameters(
                    self.edited_circuit, self.conf_,
                    self.constants_1, self.constants_2)
                self.p1, self.p2 = individual_parameters(
                    self.edited_circuit, self.parameters_,
                    self.constants_1, self.constants_2)
                # self.cov_ = np.array(json_data["Covariance"])
                # self.cov1, self.cov2 = individual_parameters(
                #     self.edited_circuit, self.cov_,
                #     self.constants_1, self.constants_2)


class NLEISCustomCircuit(BaseCircuit):
    # this class can be fully integrated into CustomCircuit in the future
    # , but for the stable performance of nleis.py, we overwrite it here
    def __init__(self, circuit='', graph=False,  **kwargs):
        """
        Constructor for a customizable nonlinear equivalent circuit model
        for NLEIS analysis.

        Parameters
        ----------
        initial_guess : numpy.ndarray
            Initial guess values for the circuit parameters.

        circuit : str
            A string representing the nonlinear equivalent circuit for NLEIS.

        graph : bool, optional
            Whether to use execution graph to process the circuit.
            Defaults to False, which uses eval based code

        Notes
        -----
        A custom NLEIS circuit is defined as a string comprised of elements
        in series (separated by a `-`),
        elements in parallel (grouped as p(x,y)),
        and elements in difference (grouped as d(x,y)).
        Each element can be appended with an integer (e.g., R0)
        or an underscore and an integer (e.g., CPE_1)
        to help keep track of multiple elements of the same type.

        Example
        -------
        A two-electrode cell with a spherical porous cathode and anode
        is represented as:

            circuit = 'd(TDSn0-TDSn1)'

        """

        super().__init__(**kwargs)
        # self.cov_ = None
        self.circuit = circuit.replace(" ", "")
        self.graph = graph
        if self.circuit:
            self.cg = CircuitGraph(self.circuit, self.constants)

        circuit_len = calculateCircuitLength(self.circuit)

        if len(self.initial_guess) + len(self.constants) != circuit_len:
            raise ValueError('The number of initial guesses ' +
                             f'({len(self.initial_guess)}) + ' +
                             'the number of constants ' +
                             f'({len(self.constants)})' +
                             ' must be equal to ' +
                             f'the circuit length ({circuit_len})')

    def fit(self, frequencies, impedance, bounds=None,
            weight_by_modulus=False, max_f=np.inf, **kwargs):
        """
        Fit the nonlinear equivalent circuit model to NLEIS data.

        Parameters
        ----------
        frequencies : numpy.ndarray
            Array of frequency values.

        impedance : numpy.ndarray, dtype=complex128
            Complex NLEIS impedance values to fit.

        bounds : 2-tuple of array_like, optional
            Lower and upper bounds on the model parameters. If not provided,
            default bounds will be used.

        weight_by_modulus : bool, optional
            If True, uses the modulus of each data point (`|Z|`)
            as the weighting factor.
            This is the standard weighting scheme when experimental variances
            are unavailable. Only applicable when `global_opt=False`.
            Default is False.

        max_f : float, optional
            The maximum frequency of interest for NLEIS.
            This is used for truncating the data based on prior experiments.
            Default is 10.

        **kwargs :
            Additional keyword arguments passed to
            `impedance.models.circuits.fitting.circuit_fit`,
            and subsequently to `scipy.optimize.curve_fit`
            or `scipy.optimize.basinhopping`.

        Raises
        ------
        TypeError
            Raised if the length of the frequency data
            does not match the NLEIS impedance data.

        ValueError
            Raised if `initial_guess` is not supplied.

        Returns
        -------
        self : object
            Returns the instance of the model for chaining.
        """
        frequencies = np.array(frequencies, dtype=float)
        impedance = np.array(impedance, dtype=complex)

        if len(frequencies) != len(impedance):
            raise TypeError('length of frequencies and impedance do not match')
        mask = np.array(frequencies) < max_f
        frequencies = frequencies[mask]
        impedance = impedance[mask]

        if self.initial_guess != []:
            parameters, conf = \
                circuit_fit(frequencies, impedance,
                            self.circuit,
                            self.initial_guess,
                            constants=self.constants,
                            bounds=bounds,
                            weight_by_modulus=weight_by_modulus,
                            graph=self.graph,
                            **kwargs)
            self.parameters_ = parameters
            if conf is not None:
                self.conf_ = conf
            # if cov is not None:
            #     self.cov_ = cov
        else:
            raise ValueError('No initial guess supplied')

        return self

    def predict(self, frequencies, max_f=np.inf, use_initial=False):
        """

        Predict impedance using an nonlinear equivalent circuit model

        Parameters
        ----------
        frequencies : numpy.ndarray
            Array of frequency values.

        max_f : float, optional
            The maximum frequency of interest for 2nd-NLEIS. Default is 10.

        use_initial : bool
            If True and the model was previously fit,
            use the initial parameters instead of the fitted ones.

        Returns
        -------
        impedance: ndarray of dtype 'complex128'
            Predicted impedance at each frequency

        """
        frequencies = np.array(frequencies, dtype=float)
        mask = np.array(frequencies) < max_f
        frequencies = frequencies[mask]
        if self.graph:
            self.cg = CircuitGraph(self.circuit, self.constants)

        if self._is_fit() and not use_initial:
            if self.graph:
                return self.cg.compute(frequencies, *self.parameters_)
            else:
                return eval(buildCircuit(self.circuit, frequencies,
                                         *self.parameters_,
                                         constants=self.constants,
                                         eval_string='',
                                         index=0)[0],
                            circuit_elements)
        else:
            warnings.warn("Simulating circuit based on initial parameters")

            if self.graph:
                return self.cg.compute(frequencies, *self.initial_guess)
            else:
                return eval(buildCircuit(self.circuit, frequencies,
                                         *self.initial_guess,
                                         constants=self.constants,
                                         eval_string='',
                                         index=0)[0],
                            circuit_elements)

    def get_param_names(self):
        """

        Converts circuit string to names and units

        """

        # parse the element names from the circuit string
        names = self.circuit.replace('d', '').replace(
            '(', '').replace(')', '')  # edited for nleis.py

        names = names.replace('p', '').replace('(', '').replace(')', '')
        names = names.replace(',', '-').replace(' ', '').split('-')

        full_names, all_units = [], []
        for name in names:
            elem = get_element_from_name(name)
            num_params = check_and_eval(elem).num_params
            units = check_and_eval(elem).units
            if num_params > 1:
                for j in range(num_params):
                    full_name = '{}_{}'.format(name, j)
                    if full_name not in self.constants.keys():
                        full_names.append(full_name)
                        all_units.append(units[j])
            else:
                if name not in self.constants.keys():
                    full_names.append(name)
                    all_units.append(units[0])

        return full_names, all_units

    def extract(self):
        """
        Extracts the parameter names and values of the fitted circuit model.

        This method retrieves the parameter names and units from the circuit
        and maps the fitted parameters to their respective names.
        If the model has been fitted,
        it also includes the fitted values in the output dictionary.

        Returns
        -------
        parameters_dict : dict
            A dictionary where the keys are parameter names,
            and the values are the corresponding fitted parameter values.
        """

        names, units = self.get_param_names()
        dict = {}
        if self._is_fit():
            params, confs = self.parameters_, self.conf_

            for name, unit, param, conf in zip(names, units, params, confs):
                dict[name] = param

        return dict

    def plot(self, ax=None, f_data=None, Z2_data=None,
             kind='nyquist', max_f=np.inf, **kwargs):
        """
        Visualizes the model and optional data as Nyquist, Bode,
        or Altair (interactive) plots.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            Axes to plot on. If None, a new figure and axes will be created.

        f_data : numpy.ndarray, optional
            Array of frequency values for input data (used for Bode plots).
            The default is None.

        Z2_data : numpy.ndarray of complex, optional
            Array of 2nd-NLEIS values (impedance data). The default is None.

        kind : {'altair', 'nyquist', 'bode'}, optional
            The type of plot to visualize.

            - 'nyquist': Nyquist plot of real vs imaginary impedance.

            - 'bode': Bode plot showing magnitude and phase.

            - 'altair': Altair plot for interactive visualizations.

            Default is 'nyquist'.

        max_f : float, optional
            The maximum frequency of interest for 2nd-NLEIS data.
            The default is 10.

        **kwargs : optional
            Additional keyword arguments passed to `matplotlib.pyplot.Line2D`
            (for 'nyquist' or 'bode') to specify properties like linewidth,
            color, marker type, etc.
            If `kind` is 'altair',
            `kwargs` is used to specify plot height as `size`.

        Returns
        -------
        ax : matplotlib.axes.Axes or tuple of Axes, optional
            Axes object(s) with the plotted data
            if 'nyquist' or 'bode' plot is used.

        chart : altair.Chart, optional
            If `kind` is 'altair', it returns Altair chart objects
            for 2nd-NLEIS data.

        Raises
        ------
        ValueError
            If an unsupported `kind` is provided.
        """

        if kind == 'nyquist':
            if ax is None:
                _, ax = plt.subplots(figsize=(5, 5))

            # we don't need the if else statement if we want
            # to enable plot without fit
            # if self._is_fit():
            if f_data is not None:
                f_pred = f_data
            else:
                f_pred = np.logspace(5, -3)

            if Z2_data is not None:
                if f_data is not None:
                    mask = np.array(f_data) < max_f
                    ax = plot_second(ax, Z2_data[mask],
                                     scale=1, fmt='s', **kwargs)
                else:
                    ax = plot_second(ax, Z2_data,
                                     scale=1, fmt='s', **kwargs)
                # impedance.py style
                # plot_nyquist(Z2_data, units='Ohms/A', ls='',
                # marker='s', ax=ax, **kwargs)

            Z2_fit = self.predict(f_pred, max_f=max_f)
            ax = plot_second(ax, Z2_fit, scale=1, fmt='-', **kwargs)
            # plot_nyquist(Z2_fit,units='Ohms/A', ls='-',
            # marker='', ax=ax, **kwargs)

            ax.legend(['Data', 'Fit'])
            return ax
        elif kind == 'bode':
            if ax is None:
                _, ax = plt.subplots(nrows=2, figsize=(8, 8))

            if f_data is not None:
                f_pred = f_data
            else:
                f_pred = np.logspace(5, -3)

            if Z2_data is not None:
                if f_data is None:
                    raise ValueError('f_data must be specified if' +
                                     ' Z_data for a Bode plot')
                mask = np.array(f_pred) < max_f
                f2 = f_data[mask]
                Z2 = Z2_data[mask]
                ax = plot_bode(f2, Z2, units='Ω/A', ls='', marker='s',
                               axes=ax, **kwargs)
            # we don't need the if else statement
            # if we want to enable plot without fit
            # if self._is_fit():
            Z2_fit = self.predict(f_pred, max_f=max_f)

            f2 = f_pred[np.array(f_pred) < max_f]

            ax = plot_bode(f2, Z2_fit, units='Ω/A', ls='-',
                           marker='o',
                           axes=ax, **kwargs)

            ax[0].set_ylabel(r'$|Z_{2}(\omega)|$ ' +
                             '$[{}]$'.format('Ω/A'), fontsize=20)
            ax[1].set_ylabel(
                r'$-\phi_{Z_{2}}(\omega)$ ' + r'$[^o]$', fontsize=20)
            ax[0].legend(['Data', 'Fit'], fontsize=20)
            ax[1].legend(['Data', 'Fit'], fontsize=20)

            return ax
        elif kind == 'altair':
            plot_dict = {}

            if (Z2_data is not None) and (f_data is not None):
                mask = np.array(f_data) < max_f
                plot_dict['data'] = {'f': f_data[mask], 'Z': Z2_data[mask]}
            # we don't need the if else statement
            # if we want to enable plot without fit
            # if self._is_fit():
            if f_data is not None:

                f_pred = f_data

            else:

                f_pred = np.logspace(5, -3)

            if self.name is not None:
                name = self.name
            else:
                name = 'fit'

            Z2_fit = self.predict(f_pred, max_f=max_f)
            mask = np.array(f_pred) < max_f
            plot_dict[name] = {'f': f_pred[mask], 'Z': Z2_fit, 'fmt': '-'}

            chart = plot_altair(plot_dict, k=2, units='Ω/A', **kwargs)

            return chart
        else:
            raise ValueError("Kind must be one of 'altair'," +
                             f"'nyquist', or 'bode' (received {kind})")

    # add on to the load function to create the graph

    def load(self, filepath, fitted_as_initial=False):
        super().load(filepath, fitted_as_initial)
        self.cg = CircuitGraph(self.circuit, self.constants)



import numpy as np
from impedance.preprocessing import ignoreBelowX
from impedance.tests.test_preprocessing import frequencies \
    as example_frequencies
from impedance.tests.test_preprocessing import Z_correct
from nleis.fitting import buildCircuit, \
    circuit_fit, mape, mae, extract_circuit_elements, \
    set_default_bounds, seq_fit_param
from nleis.nleis_fitting import data_processing, \
    simul_fit, individual_parameters
import os
import pytest


test_dir = os.path.dirname(os.path.abspath(__file__))
data_dir = os.path.join(test_dir, '../data')


def test_data_processing():

    # # get example data
    # # The example is shown in "Getting Started" page

    frequencies = np.loadtxt(os.path.join(data_dir, 'freq_30a.txt'))
    Z1 = np.loadtxt(os.path.join(data_dir, 'Z1s_30a.txt')).view(complex)[1]
    Z2 = np.loadtxt(os.path.join(data_dir, 'Z2s_30a.txt')).view(complex)[1]

    max_f = 10

    f, Z1, Z2, f2_truncated, Z2_truncated = data_processing(
        frequencies, Z1, Z2, max_f=max_f)
    assert len(f) == len(Z1) == len(Z2)
    assert Z1.imag.max() < 0
    assert f2_truncated.max() < max_f
    assert len(f2_truncated) == len(Z2_truncated)


def test_set_default_bounds():
    # Test example circuit that calculate the difference between
    # two porous electrode with spherical particle
    # and with platelet like particle
    circuit = 'd(TDSn0,TDPn1)'

    # Test with no constants
    default_bounds = (np.zeros(14), np.inf*np.ones(14))
    default_bounds[0][5] = -np.inf
    default_bounds[0][6] = -0.5
    default_bounds[1][6] = 0.5
    default_bounds[0][12] = -np.inf
    default_bounds[0][13] = -0.5
    default_bounds[1][13] = 0.5

    bounds_from_func = set_default_bounds(circuit)

    assert np.allclose(default_bounds, bounds_from_func)

    # Test with constants
    constants = {'TDSn0_0': 1}
    default_bounds = (np.zeros(13), np.inf*np.ones(13))
    default_bounds[0][4] = -np.inf
    default_bounds[0][5] = -0.5
    default_bounds[1][5] = 0.5
    default_bounds[0][11] = -np.inf
    default_bounds[0][12] = -0.5
    default_bounds[1][12] = 0.5
    bounds_from_func = set_default_bounds(circuit, constants=constants)

    assert np.allclose(default_bounds, bounds_from_func)

    circuit = 'CPE0'
    default_bounds = (np.zeros(2), np.inf*np.ones(2))
    default_bounds[1][1] = 1
    bounds_from_func = set_default_bounds(circuit, constants=constants)

    assert np.allclose(default_bounds, bounds_from_func)

    circuit = 'La0'
    default_bounds = (np.zeros(2), np.inf*np.ones(2))
    default_bounds[1][1] = 1
    bounds_from_func = set_default_bounds(circuit, constants=constants)

    assert np.allclose(default_bounds, bounds_from_func)

    circuit = 'TPn0'
    default_bounds = (np.zeros(4), np.inf*np.ones(4))
    default_bounds[0][3] = -0.5
    default_bounds[1][3] = 0.5
    bounds_from_func = set_default_bounds(circuit, constants=constants)

    assert np.allclose(default_bounds, bounds_from_func)

    circuit = 'RCn0'
    default_bounds = (np.zeros(3), np.inf*np.ones(3))
    default_bounds[0][2] = -0.5
    default_bounds[1][2] = 0.5
    bounds_from_func = set_default_bounds(circuit, constants=constants)

    assert np.allclose(default_bounds, bounds_from_func)

    circuit = 'RCDn0'
    bounds_from_func = set_default_bounds(circuit, constants=constants)
    default_bounds = (np.zeros(6), np.inf*np.ones(6))
    default_bounds[0][5] = -0.5
    default_bounds[1][5] = 0.5
    default_bounds[0][4] = -np.inf

    assert np.allclose(default_bounds, bounds_from_func)

    circuit = 'RCSn0'
    bounds_from_func = set_default_bounds(circuit, constants=constants)
    assert np.allclose(default_bounds, bounds_from_func)

    circuit = 'TLMSn'
    bounds_from_func = set_default_bounds(circuit, constants=constants)
    default_bounds = (np.zeros(11), np.inf*np.ones(11))
    default_bounds[0][9] = -0.5
    default_bounds[0][10] = -0.5
    default_bounds[1][9] = 0.5
    default_bounds[1][10] = 0.5
    default_bounds[0][8] = -np.inf
    assert np.allclose(default_bounds, bounds_from_func)

    circuit = 'TLMDn'
    bounds_from_func = set_default_bounds(circuit, constants=constants)
    assert np.allclose(default_bounds, bounds_from_func)

    circuit = 'TLMn'
    bounds_from_func = set_default_bounds(circuit, constants=constants)
    default_bounds = (np.zeros(8), np.inf*np.ones(8))
    default_bounds[0][6] = -0.5
    default_bounds[0][7] = -0.5
    default_bounds[1][6] = 0.5
    default_bounds[1][7] = 0.5
    assert np.allclose(default_bounds, bounds_from_func)

    circuit = 'RCSQ'
    bounds_from_func = set_default_bounds(circuit, constants=constants)
    default_bounds = (np.zeros(5), np.inf*np.ones(5))
    default_bounds[1][2] = 1
    assert np.allclose(default_bounds, bounds_from_func)


# This test remain unchanged as we are adopting it from impedance.py
def test_circuit_fit():

    # Test trivial model (10 Ohm resistor)
    circuit = 'R0'
    initial_guess = [10]

    results_simple = [10]

    frequencies = np.array([10, 100, 1000])
    Z_data = np.array([10, 10, 10])  # impedance is real

    assert np.allclose(circuit_fit(frequencies, Z_data, circuit,
                                   initial_guess, constants={},
                                   global_opt=True)[0],
                       results_simple, rtol=1e-1)

    # check that list inputs work
    frequency_list = [10, 100, 1000]
    Z_data_list = [10, 10, 10]

    assert np.allclose(circuit_fit(frequency_list, Z_data_list, circuit,
                                   initial_guess, constants={},
                                   global_opt=True)[0],
                       results_simple, rtol=1e-1)

    # Test example circuit from "Getting Started" page
    circuit = 'R0-p(R1,C1)-p(R2-Wo1,C2)'
    initial_guess = [.01, .01, 100, .01, .05, 100, 1]
    bounds = [(0, 0, 0, 0, 0, 0, 0),
              (10, 1, 1e3, 1, 1, 1e4, 100)]

    # results change slightly using predefined bounds
    results_local = np.array([1.65e-2, 8.68e-3, 3.32, 5.39e-3,
                              6.31e-2, 2.33e2, 2.20e-1])
    results_local_bounds = results_local.copy()
    results_local_bounds[5] = 2.38e2
    results_local_weighted = np.array([1.64e-2, 9.06e-3, 3.06,
                                       5.29e-3, 1.45e-1, 1.32e3, 2.02e-1])

    results_global = np.array([1.65e-2, 5.34e-3, 0.22, 9.15e-3,
                               1.31e-1, 1.10e3, 2.78])

    # Filter
    example_frequencies_filtered, \
        Z_correct_filtered = ignoreBelowX(example_frequencies, Z_correct)

    # Test local fitting
    assert np.allclose(circuit_fit(example_frequencies_filtered,
                                   Z_correct_filtered, circuit,
                                   initial_guess, constants={})[0],
                       results_local, rtol=1e-2)

    # Test local fitting with predefined bounds
    assert np.allclose(circuit_fit(example_frequencies_filtered,
                                   Z_correct_filtered, circuit,
                                   initial_guess, bounds=bounds,
                                   constants={})[0],
                       results_local_bounds, rtol=1e-2)

    # Test local fitting with predefined weights
    # Use abs(Z), stacked in order of (Re, Im) components
    sigma = np.hstack((np.abs(Z_correct_filtered),
                       np.abs(Z_correct_filtered)))
    assert np.allclose(circuit_fit(example_frequencies_filtered,
                                   Z_correct_filtered, circuit,
                                   initial_guess, sigma=sigma,
                                   constants={})[0],
                       results_local_weighted, rtol=1e-2)

    # Test if using weight_by_modulus=True produces the same results
    assert np.allclose(circuit_fit(example_frequencies_filtered,
                                   Z_correct_filtered, circuit,
                                   initial_guess, weight_by_modulus=True,
                                   constants={})[0],
                       results_local_weighted, rtol=1e-2)

    # Test global fitting on multiple seeds
    # All seeds should converge to the same parameter values
    # seed = 0 (default)
    assert np.allclose(circuit_fit(example_frequencies_filtered,
                                   Z_correct_filtered, circuit,
                                   initial_guess, constants={},
                                   global_opt=True)[0],
                       results_global, rtol=1e-1)

    # seed = 0, with predefined bounds
    assert np.allclose(circuit_fit(example_frequencies_filtered,
                                   Z_correct_filtered, circuit,
                                   initial_guess, constants={},
                                   global_opt=True, bounds=bounds,
                                   seed=0)[0],
                       results_global, rtol=1e-1)

    # seed = 1
    assert np.allclose(circuit_fit(example_frequencies_filtered,
                                   Z_correct_filtered, circuit,
                                   initial_guess, constants={},
                                   global_opt=True, seed=1)[0],
                       results_global, rtol=1e-1)

    # seed = 42
    assert np.allclose(circuit_fit(example_frequencies_filtered,
                                   Z_correct_filtered, circuit,
                                   initial_guess, constants={},
                                   global_opt=True, seed=42)[0],
                       results_global, rtol=1e-1)


def test_buildCircuit():

    # Test a simple porous electrode with contact resistor and inductance
    circuit = 'L0-R0-TDS0'
    params = [1e-5, 0.01, 1, 0.1, 10, 1, 1000]
    frequencies = [1000.0, 5.0, 0.01]

    assert buildCircuit(circuit, frequencies, *params,
                        constants={})[0].replace(' ', '') == \
        's([L([1e-05],[1000.0,5.0,0.01]),' + \
        'R([0.01],[1000.0,5.0,0.01]),' + \
        'TDS([1.0,0.1,10.0,1.0,1000.0],[1000.0,5.0,0.01])])'

# Test the substraction between two elements
    circuit = 'd(R1,R2)'
    params = [0.1, 0.2]
    frequencies = [1000.0, 5.0, 0.01]

    assert buildCircuit(circuit, frequencies, *params,
                        constants={})[0].replace(' ', '') == \
        'd([R([0.1],[1000.0,5.0,0.01]),' + \
        'R([0.2],[1000.0,5.0,0.01])])'

    # Test nested parallel and difference groups
    circuit = 'R0-d(p(R1, C1)-R2, C2)'
    params = [1, 2, 3, 4, 5]
    frequencies = [1000.0, 5.0, 0.01]

    assert buildCircuit(circuit, frequencies, *params,
                        constants={})[0].replace(' ', '') == \
        's([R([1],[1000.0,5.0,0.01]),' + \
        'd([s([p([R([2],[1000.0,5.0,0.01]),' + \
        'C([3],[1000.0,5.0,0.01])]),' + \
        'R([4],[1000.0,5.0,0.01])]),' + \
        'C([5],[1000.0,5.0,0.01])])])'

    # Test single element circuit
    circuit = 'R1'
    params = [100]
    frequencies = [1000.0, 5.0, 0.01]

    assert buildCircuit(circuit, frequencies, *params,
                        constants={})[0].replace(' ', '') == \
        'R([100],[1000.0,5.0,0.01])'

    # Test two elements circuit with constants for one elements
    circuit = 'R1-R2'
    params = [100]
    frequencies = [1000.0, 5.0, 0.01]

    assert buildCircuit(circuit, frequencies, *params,
                        constants={'R2': 1})[0].replace(' ', '') == \
        's([R([100],[1000.0,5.0,0.01]),' +\
        'R([1],[1000.0,5.0,0.01])])'

    # Test constants in circuit
    circuit = 'Wo1'
    params = [100]
    frequencies = [1000.0, 5.0, 0.01]
    constants = {'Wo1_1': 1}

    assert buildCircuit(circuit, frequencies, *params,
                        constants=constants)[0].replace(' ', '') == \
        'Wo([100,1],[1000.0,5.0,0.01])'


def test_mae():
    a = np.array([2 + 4*1j, 3 + 2*1j])
    b = np.array([2 + 4*1j, 3 + 2*1j])

    assert mae(a, b) == 0.0

    c = np.array([2 + 4*1j, 1 + 4*1j])
    d = np.array([4 + 2*1j, 3 + 2*1j])
    assert np.isclose(mae(c, d), np.sqrt(8))


def test_mape():
    a = np.array([2 + 4*1j, 3 + 2*1j])
    b = np.array([2 + 4*1j, 3 + 2*1j])

    assert mape(a, b) == 0.0

    c = np.array([2 + 4*1j, 1 + 4*1j])
    d = np.array([4 + 2*1j, 3 + 2*1j])
    assert np.isclose(mape(c, d), 0.5*np.sqrt(8) *
                      (1/np.sqrt(20)+1/np.sqrt(17))*100)


def test_element_extraction():
    circuit = 'd(TDSn0,TDPn1)'
    extracted_elements = extract_circuit_elements(circuit)
    assert extracted_elements == ['TDSn0', 'TDPn1']


def test_seq_fit_param():
    input_dict = {'TDS0_1': 0, 'TDS1_3': 0}
    target_arr = ['TDS0', 'TDS1']
    output_arr = ['TDSn0', 'TDSn1']
    output_dict = seq_fit_param(input_dict, target_arr, output_arr)
    assert output_dict == {'TDSn0_1': 0, 'TDSn1_3': 0}

    with pytest.raises(ValueError):
        output_arr = ['TDSn0', 'TDSn1', 'TDSn2']
        output_dict = seq_fit_param(input_dict, target_arr, output_arr)


def test_simul_fit():

    # Test example shown in  "Getting Started" page
    # Note: solution to the solid state diffuion might
    # converge to different value given the exact
    # envrionment used (i.e. large confidence interval)

    frequencies = np.loadtxt(os.path.join(data_dir, 'freq_30a.txt'))
    Z1 = np.loadtxt(os.path.join(data_dir, 'Z1s_30a.txt')).view(complex)[1]
    Z2 = np.loadtxt(os.path.join(data_dir, 'Z2s_30a.txt')).view(complex)[1]

    f, Z1, Z2, f2_trunc, Z2_trunc = data_processing(frequencies, Z1, Z2)

    circ_str_1 = 'L0-R0-TDS0-TDS1'
    circ_str_2 = 'd(TDSn0,TDSn1)'
    edited_circuit = 'L0-R0-TDSn0-TDSn1'

    # The test framework here can be improved in the future
    # with better and stable initial guess
    results = [9.81368514e-08, 1.34551972e-02, 2.52387276e-02, 5.06176242e-03,
               8.82244297e+00, 8.70692162e-05, 3.55536976e+00, 1.22576118e+01,
               8.75169434e-02, 2.09045802e-02, 1.13804384e-03, 8.13658287e-01,
               1.83783329e+02, 3.20554700e+09, 1.02277512e+00, 6.39228801e-03]

    initial_guess = results

    # initial_guess
    # L0,RO
    # TDS0 + additioal nonlinear parameters
    # TDS1 + additioal nonlinear parameters

    p, perror = simul_fit(f, Z1, Z2, circ_str_1, circ_str_2,
                          edited_circuit, initial_guess, constants_1={},
                          constants_2={},
                          bounds=None, opt='max', cost=0.5, max_f=10,
                          param_norm=False, positive=False)

    assert np.allclose(p, results, rtol=1e-3, atol=1e-3)

    # test the option of parameter normalization
    with pytest.warns(UserWarning, match="inf is detected in the bounds"):
        bounds = set_default_bounds(edited_circuit)
        p, perror = simul_fit(f, Z1, Z2, circ_str_1, circ_str_2,
                              edited_circuit, initial_guess, constants_1={},
                              constants_2={},
                              bounds=bounds, opt='max', cost=0.5, max_f=10,
                              param_norm=True, positive=True)


def test_individual_parameters():

    # EIS and 2nd-NLEIS circ_str are defined here to make sure
    # we know we have a a pair of consistent models

    circ_str_1 = 'L0-R0-TDS0-TDS1'  # noqa: F841
    circ_str_2 = 'd(TDSn0,TDSn1)'  # noqa: F841
    circ_str = 'L0-R0-TDSn0-TDSn1'

    initial_guess = [1e-7, 1e-3,  # L0,RO
                     5e-3, 1e-3, 10, 1e-2, 100, 10, 0.1,
                     # TDS0 + additioal nonlinear parameters
                     1e-3, 1e-3, 1e-3, 1e-2, 1000, 0, 0,
                     # TDS1 + additioal nonlinear parameters
                     ]
    # Test without constant
    p1, p2 = individual_parameters(
        circ_str, initial_guess, constants_1={}, constants_2={})
    assert np.allclose(p1, [1e-7, 1e-3, 5e-3, 1e-3, 10.0,
                       1e-02, 100.0, 1e-3, 1e-03, 1e-03, 1e-02, 1000.0])
    assert np.allclose(p2, [5e-3, 1e-3, 10, 1e-2, 100,
                       10, 0.1, 1e-3, 1e-3, 1e-3, 1e-2, 1000, 0, 0])

    # Test with constant
    initial_guess = [1e-7, 1e-3,  # L0,RO
                     5e-3, 10, 1e-2, 100, 10, 0.1,
                     # TDS0 + additioal nonlinear parameters
                     1e-3, 1e-3, 1e-3, 1e-2, 1000, 0
                     # TDS1 + additioal nonlinear parameters
                     ]
    p1, p2 = individual_parameters(circ_str, initial_guess, constants_1={
                                   'TDS0_1': 0}, constants_2={'TDSn1_6': 0})

    assert np.allclose(p1, [1e-7, 1e-3, 5e-3, 10, 1e-02,
                       100.0, 1e-3, 1e-03, 1e-03, 1e-02, 1000.0])
    assert np.allclose(p2, [5e-3, 10, 1e-2, 100, 10,
                       0.1, 1e-3, 1e-3, 1e-3, 1e-2, 1000, 0])

    # test the case when only one electrode is contributing to the 2nd-NLEIS
    # without constnats
    circ_str_1 = 'L0-R0-TDS0-TDS1'  # noqa: F841
    circ_str_2 = 'TDSn0'  # noqa: F841
    circ_str = 'L0-R0-TDSn0-TDS1'

    initial_guess = [1e-7, 1e-3,  # L0,RO
                     5e-3, 1e-3, 10, 1e-2, 100, 10, 0.1,
                     # TDS0 + additioal nonlinear parameters
                     1e-3, 1e-3, 1e-3, 1e-2, 1000, 0, 0,
                     # TDS1 + additioal nonlinear parameters
                     ]

    p1, p2 = individual_parameters(
        circ_str, initial_guess, constants_1={}, constants_2={})
    assert np.allclose(p1, [1e-7, 1e-3, 5e-3, 1e-3, 10.0,
                       1e-02, 100.0, 1e-3, 1e-03, 1e-03, 1e-02, 1000.0])

    assert np.allclose(p2, [5e-3, 1e-3, 10, 1e-2, 100, 10, 0.1])

    # test the case when only one electrode is contributing to the 2nd-NLEIS
    # with constants
    circ_str_1 = 'L0-R0-TDS0-TDS1'  # noqa: F841
    circ_str_2 = 'TDSn0'  # noqa: F841
    circ_str = 'L0-R0-TDSn0-TDS1'

    initial_guess = [1e-7, 1e-3,  # L0,RO
                     5e-3, 1e-3, 10, 1e-2, 100, 10,
                     # TDS0 + additioal nonlinear parameters
                     1e-3, 1e-3, 1e-2, 1000
                     # TDS1 + additioal nonlinear parameters
                     ]

    p1, p2 = individual_parameters(
        circ_str, initial_guess,
        constants_1={'TDS1_0': 0}, constants_2={'TDSn0_6': 0.1})
    assert np.allclose(p1, [1e-7, 1e-3, 5e-3, 1e-3, 10.0,
                       1e-02, 100.0, 1e-3, 1e-03, 1e-02, 1000.0])
    assert np.allclose(p2, [5e-3, 1e-3, 10, 1e-2, 100, 10])

    # test the case when only one electrode is contributing to the 2nd-NLEIS
    # with constants + L0 as constants
    circ_str_1 = 'L0-R0-TDS0-TDS1'  # noqa: F841
    circ_str_2 = 'TDSn0'  # noqa: F841
    circ_str = 'L0-R0-TDSn0-TDS1'

    initial_guess = [1e-3,  # L0,RO
                     5e-3, 1e-3, 10, 1e-2, 100, 10,
                     # TDS0 + additioal nonlinear parameters
                     1e-3, 1e-3, 1e-2, 1000
                     # TDS1 + additioal nonlinear parameters
                     ]

    p1, p2 = individual_parameters(
        circ_str, initial_guess,
        constants_1={'L0': 1e-7, 'TDS1_0': 0}, constants_2={'TDSn0_6': 0.1})
    assert np.allclose(p1, [1e-3, 5e-3, 1e-3, 10.0,
                       1e-02, 100.0, 1e-3, 1e-03, 1e-02, 1000.0])

    assert np.allclose(p2, [5e-3, 1e-3, 10, 1e-2, 100, 10])



# import warnings
# from scipy.linalg import inv
# from scipy.optimize import basinhopping

import numpy as np
from scipy.optimize import curve_fit
from impedance.models.circuits.elements import circuit_elements, \
    get_element_from_name
from impedance.models.circuits.fitting import check_and_eval
from .fitting import set_default_bounds, buildCircuit, extract_circuit_elements
from scipy.optimize import minimize
import warnings
from nleis.fitting import CircuitGraph

# Customize warning format (here, simpler and just the message)
warnings.formatwarning = lambda message, category, filename, lineno, \
    line=None: f'{category.__name__}: {message}\n'

ints = '0123456789'


def data_processing(f, Z1, Z2, max_f=10):
    """
    Simple data processing function for EIS and 2nd-NLEIS simultaneously.

    This function processes frequency (`f`), EIS data (`Z1`), and 2nd-NLEIS
    data (`Z2`), removing high-frequency inductance effects and truncating the
    data for 2nd-NLEIS analysis based on the specified maximum frequency.

    Parameters
    ----------
    f : numpy.ndarray
        Array of frequency values.

    Z1 : numpy.ndarray of dtype complex128
        EIS data (complex impedance values).

    Z2 : numpy.ndarray of dtype complex128
        2nd-NLEIS data (complex impedance values).

    max_f : float, optional
        The maximum frequency of interest for 2nd-NLEIS. Frequencies higher
        than `max_f` will be truncated for 2nd-NLEIS data. Default is 10.

    Returns
    -------
    f : numpy.ndarray
        Frequencies after removing high-frequency inductance effects.

    Z1 : numpy.ndarray of dtype complex128
        EIS data with high-frequency inductance removed.

    Z2 : numpy.ndarray of dtype complex128
        2nd-NLEIS data that matches the length of
        high-frequency inductance removed EIS.

    f2_truncated : numpy.ndarray
        Frequencies that are less than
        the specified maximum frequency (`max_f`)
        for 2nd-NLEIS, while removes high frequency inductance

    Z2_truncated : numpy.ndarray of dtype complex128
        2nd-NLEIS data corresponding to the truncated frequency range.

    """

    mask = np.array(Z1.imag) < 0
    f = f[mask]
    Z1 = Z1[mask]
    Z2 = Z2[mask]
    mask1 = np.array(f) < max_f
    f2_truncated = f[mask1]
    Z2_truncated = Z2[mask1]
    return (f, Z1, Z2, f2_truncated, Z2_truncated)


def simul_fit(frequencies, Z1, Z2, circuit_1, circuit_2, edited_circuit,
              initial_guess, constants_1={}, constants_2={},
              bounds=None, opt='max', cost=0.5, max_f=10, param_norm=True,
              positive=True, graph=False, **kwargs):
    """
    Main function for the simultaneous fitting of EIS and 2nd-NLEIS data.

    This function fits the equivalent circuits for both EIS and 2nd-NLEIS data
    simultaneously using `scipy.optimize.curve_fit
    <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.curve_fit.html>`_
    by default. It supports
    parameter normalization, bounds customization, and two optimization methods
    (`'max'` for maximum normalization and `'neg'` for negative log-likelihood)

    Parameters
    ----------
    frequencies : numpy.ndarray
        Array of frequency values.

    Z1 : numpy.ndarray of dtype complex128
        EIS data (complex impedance values).

    Z2 : numpy.ndarray of dtype complex128
        2nd-NLEIS data (complex impedance values).

    circuit_1 : str
        String defining the EIS equivalent circuit to be fit.

    circuit_2 : str
        String defining the 2nd-NLEIS equivalent circuit to be fit.

    edited_circuit : str
        Edited circuit string that is applied to both EIS and NLEIS fitting.

        Example:

        circuit_1 = L0-R0-TDP0-TDS1

        circuit_2 = TDPn0-TDSn1

        edited_circuit = L0-R0-TDPn0-TDSn1

    initial_guess : list of float
        Initial guesses for the fit parameters.

    constants_1 : dict, optional
        Parameters and their values to hold constant during EIS fitting.
        Defaults to an empty dictionary.

    constants_2 : dict, optional
        Parameters and their values to hold constant during 2nd-NLEIS fitting.
        Defaults to an empty dictionary.

    bounds : 2-tuple of array_like, optional
        Lower and upper bounds on parameters. If not provided, default bounds
        will be set based on the circuit elements.

        Special bounds:

        - CPE alpha has an upper bound of 1.

        - Symmetry parameter (ε) for 2nd-NLEIS has bounds between -0.5 and 0.5.

        - Curvature parameter (κ) for 2nd-NLEIS has bounds
          between -np.inf and np.inf.

    opt : str, optional
        Optimization method. Default is `'max'` for maximum normalization.
        Negative log-likelihood is also supported as `'neg'`.

    cost : float, optional
        Weighting between EIS and 2nd-NLEIS data. A value greater than 0.5
        puts more weight on EIS,
        while a value less than 0.5 puts more weight on 2nd-NLEIS.
        Default is 0.5.

    max_f : int, optional
        Maximum frequency of interest for 2nd-NLEIS. Default is 10.

    param_norm : bool, optional
        If True, parameter normalization is applied to improve convergence.
        Defaults to True.

    positive : bool, optional
        If True, high-frequency inductance is eliminated. Defaults to True.

    graph : bool, optional
        Whether to use execution graph to process the circuit.
        Defaults to False, which uses eval based code

    kwargs :
        Additional keyword arguments passed to `scipy.optimize.curve_fit`.

    Returns
    -------
    p_values : list of float
        Best-fit parameters for EIS and 2nd-NLEIS data.

    p_errors : list of float
        One standard deviation error estimates for fitting parameters.
        If using `'neg'` optimization, this value will be None.

    """
    # Todo improve the the negtive loglikelihood,
    # the code works fine for RC but not porous electrode

    # set upper and lower bounds on a per-element basis
    if bounds is None:
        combined_constant = constants_2.copy()
        combined_constant.update(constants_1)
        bounds = set_default_bounds(
            edited_circuit, constants=combined_constant)
        ub = np.ones(len(bounds[1]))
    else:
        if param_norm:
            inf_in_bounds = np.any(np.isinf(bounds[0])) \
                or np.any(np.isinf(bounds[1]))
            if inf_in_bounds:
                lb = np.where(np.array(bounds[0]) == -np.inf, -1e10, bounds[0])
                ub = np.where(np.array(bounds[1]) == np.inf, 1e10, bounds[1])
                bounds = (lb, ub)
                warnings.warn("inf is detected in the bounds, "
                              "to enable parameter normalization, "
                              "the bounds has been capped at 1e10. "
                              "You can disable parameter normalization "
                              "by set param_norm to False .")
            else:
                ub = bounds[1]

            bounds = bounds/ub
        else:
            ub = np.ones(len(bounds[1]))

    initial_guess = initial_guess/ub

    if positive:
        mask1 = np.array(Z1.imag) < 0
        frequencies = frequencies[mask1]
        Z1 = Z1[mask1]
        Z2 = Z2[mask1]
        mask2 = np.array(frequencies) < max_f
        Z2 = Z2[mask2]
    else:
        mask2 = np.array(frequencies) < max_f
        Z2 = Z2[mask2]

    Z1stack = np.hstack([Z1.real, Z1.imag])
    Z2stack = np.hstack([Z2.real, Z2.imag])
    Zstack = np.hstack([Z1stack, Z2stack])
    # weighting scheme for fitting
    if opt == 'max':
        if 'maxfev' not in kwargs:
            kwargs['maxfev'] = 1e5
        if 'ftol' not in kwargs:
            kwargs['ftol'] = 1e-13
        Z1max = max(np.abs(Z1))
        Z2max = max(np.abs(Z2))

        sigma1 = np.ones(len(Z1stack))*Z1max/(cost**0.5)
        sigma2 = np.ones(len(Z2stack))*Z2max/((1-cost)**0.5)
        kwargs['sigma'] = np.hstack([sigma1, sigma2])

        popt, pcov = curve_fit(
            wrapCircuit_simul(edited_circuit, circuit_1, constants_1,
                              circuit_2, constants_2,
                              ub, max_f, graph=graph), frequencies, Zstack,
            p0=initial_guess, bounds=bounds, **kwargs)

    # Calculate one standard deviation error estimates for fit parameters,
    # defined as the square root of the diagonal of the covariance matrix.
    # https://stackoverflow.com/a/52275674/5144795
    # and the following for the bounded and normalized case
    # https://stackoverflow.com/questions/14854339/in-scipy-how-and-why-does-curve-fit-calculate-the-covariance-of-the-parameter-es
        perror = np.sqrt(np.diag(ub*pcov*ub.T))

        return popt*ub, perror
    if opt == 'neg':
        # This method does not provides converge solution
        # for porous electrode model under current development
        # The method is introduced by Kirk et al, J. Electrochem. Soc. (2023)
        # doi: 10.1149/1945-7111/acada7
        # to support analysis for single particle model

        bounds = tuple(tuple((bounds[0][i], bounds[1][i]))
                       for i in range(len(bounds[0])))

        res = minimize(
            wrapNeg_log_likelihood(frequencies, Z1, Z2, edited_circuit,
                                   circuit_1, constants_1,
                                   circuit_2, constants_2,
                                   ub, max_f, cost=cost, graph=graph),
            x0=initial_guess, bounds=bounds, **kwargs)

        return (res.x*ub, None)


def wrapNeg_log_likelihood(frequencies, Z1, Z2, edited_circuit,
                           circuit_1, constants_1,
                           circuit_2, constants_2, ub, max_f=10, cost=0.5,
                           graph=False):
    ''' wraps function so we can pass the circuit string
    for negtive log likelihood optimization'''

    def wrappedNeg_log_likelihood(parameters):
        """ returns a stacked array of real and imaginary impedance
        components

        Parameters
        ----------
        frequencies: list of floats
        Z1: EIS data
        Z2: NLEIS data
        edited_circuit : string
        circuit_1 : string
        constants_1 : dict
        circuit_2 : string
        constants_2 : dict
        ub : list of floats upper bound if bounds are provided
        max_f: int
        cost : float, optional
            Weighting between EIS and 2nd-NLEIS data. A value greater than 0.5
            puts more weight on EIS,
            while a value less than 0.5 puts more weight on 2nd-NLEIS.
            Default is 0.5.
        graph : bool, optional
            Whether to use execution graph to process the circuit.
            Defaults to False, which uses eval based code
        parameters : list of floats

        Returns
        -------
        array of floats

        """
        f1 = frequencies
        mask = np.array(frequencies) < max_f
        f2 = frequencies[mask]
        x1, x2 = wrappedImpedance(edited_circuit,
                                  circuit_1, constants_1, circuit_2,
                                  constants_2, f1, f2, parameters*ub,
                                  graph=graph)

        # No normalization in currently applied
        # Z1max = max(np.abs(Z1))
        # Z2max = max(np.abs(Z2))
        # log1 = np.log(sum(((Z1.real-x1.real)/Z1max)**2))
        # +np.log(sum(((Z1.imag-x1.imag)/Z1max)**2))
        # log2 = np.log(sum(((Z2.real-x2.real)/Z2max)**2))
        # +np.log(sum(((Z2.imag-x2.imag)/Z2max)**2))

        log1 = np.log(sum(abs(Z1-x1)**2))
        log2 = np.log(sum(abs(Z2-x2)**2))
        return (cost*log1+(1-cost)*log2)
    return wrappedNeg_log_likelihood


def wrapCircuit_simul(edited_circuit, circuit_1, constants_1, circuit_2,
                      constants_2, ub, max_f=10, graph=False):
    """ wraps function so we can pass the circuit string
    for simultaneous fitting """
    def wrappedCircuit_simul(frequencies, *parameters):
        """ returns a stacked array of real and imaginary impedance
        components

        Parameters
        ----------
        edited_circuit : string
        circuit_1 : string
        constants_1 : dict
        circuit_2 : string
        constants_2 : dict
        ub : list of floats upper bound if bounds are provided
        max_f: float
        graph : bool, optional
            Whether to use execution graph to process the circuit.
            Defaults to False, which uses eval based code

        frequencies : list of floats

        parameters : list of floats


        Returns
        -------
        array of floats

        """

        f1 = frequencies
        mask = np.array(frequencies) < max_f
        f2 = frequencies[mask]
        x1, x2 = wrappedImpedance(edited_circuit,
                                  circuit_1, constants_1,
                                  circuit_2, constants_2,
                                  f1, f2, parameters*ub, graph=graph)

        y1_real = np.real(x1)
        y1_imag = np.imag(x1)
        y1_stack = np.hstack([y1_real, y1_imag])
        y2_real = np.real(x2)
        y2_imag = np.imag(x2)
        y2_stack = np.hstack([y2_real, y2_imag])

        return np.hstack([y1_stack, y2_stack])
    return wrappedCircuit_simul


def wrappedImpedance(edited_circuit, circuit_1, constants_1, circuit_2,
                     constants_2, f1, f2, parameters, graph=False):
    """
    Calculate EIS and 2nd-NLEIS impedances using the provided circuits.

    This function evaluates the equivalent circuit models for both EIS and
    2nd-NLEIS using the provided frequencies and parameters.

    Parameters
    ----------
    edited_circuit : str
        The edited circuit string which contains
        the combined elements of EIS and 2nd-NLEIS.

    circuit_1 : str
        The equivalent circuit string for EIS.

    constants_1 : dict
        Constants used in the EIS circuit model.

    circuit_2 : str
        The equivalent circuit string for 2nd-NLEIS.

    constants_2 : dict
        Constants used in the 2nd-NLEIS circuit model.

    f1 : list of float
        frequencies  for EIS.

    f2 : list of float
        frequencies for 2nd-NLEIS.

    parameters : list of float
        Full set of parameters derived from the edited circuit string.
    graph : bool, optional
        Whether to use execution graph to process the circuit.
        Defaults to False, which uses eval based code

    Returns
    -------
    x1 : numpy.ndarray
        Calculated impedance for EIS (Z1).

    x2 : numpy.ndarray
        Calculated impedance for 2nd-NLEIS (Z2).
    """

    p1, p2 = individual_parameters(
        edited_circuit, parameters, constants_1, constants_2)
    if graph:
        graph_EIS = CircuitGraph(circuit_1, constants_1)
        graph_NLEIS = CircuitGraph(circuit_2, constants_2)
        x1 = graph_EIS.compute(f1, *p1)
        x2 = graph_NLEIS.compute(f2, *p2)
    else:
        x1 = eval(buildCircuit(circuit_1, f1, *p1,
                               constants=constants_1, eval_string='',
                               index=0)[0],
                  circuit_elements)
        x2 = eval(buildCircuit(circuit_2, f2, *p2,
                               constants=constants_2, eval_string='',
                               index=0)[0],
                  circuit_elements)
    return (x1, x2)


def individual_parameters(edited_circuit,
                          parameters, constants_1, constants_2):
    """
    Separate parameters for EIS and 2nd-NLEIS based on the edited circuit.

    This function parses the edited circuit string and assigns the parameters
    to either EIS or 2nd-NLEIS depending on the element type.

    Parameters
    ----------
    edited_circuit : str
        Edited circuit string combining both EIS and 2nd-NLEIS elements.
        For example, if the EIS string is 'L0-R0-TDS0-TDS1' and the 2nd-NLEIS
        string is 'd(TDSn0-TDSn1)',
        the edited string becomes 'L0-R0-TDSn0-TDSn1'.

    parameters : list of float
        Full set of parameters corresponding to the edited circuit string.

    constants_1 : dict
        Constants for the EIS circuit elements.

    constants_2 : dict
        Constants for the 2nd-NLEIS circuit elements.

    Returns
    -------
    p1 : list of float
        Parameters for EIS.

    p2 : list of float
        Parameters for 2nd-NLEIS.
    """

    if edited_circuit == '':
        return [], []
    parameters = list(parameters)
    elements_1 = extract_circuit_elements(edited_circuit)
    p1 = []
    p2 = []
    index = 0
    # Parse elements and store values
    for elem in elements_1:

        raw_elem = get_element_from_name(elem)
        nleis_elem_number = check_and_eval(raw_elem).num_params
        # this might be improvable, but depends on
        # how we want to define the name
        if (elem[0] == 'T' or elem[0:2] == 'RC') and 'n' in elem:
            # check for nonlinear element
            eis_elem_number = check_and_eval(raw_elem[0:-1]).num_params

        else:
            eis_elem_number = nleis_elem_number

        for j in range(nleis_elem_number):
            if eis_elem_number > 1:
                if j < eis_elem_number:
                    eis_current_elem = elem.replace("n", "") + '_{}'.format(j)
                else:
                    eis_current_elem = None
                nleis_current_elem = elem + '_{}'.format(j)

            else:
                eis_current_elem = elem
                nleis_current_elem = elem
            if eis_current_elem in constants_1.keys():
                continue
            elif (nleis_current_elem in constants_2.keys() and
                  eis_current_elem not in constants_1.keys()):
                continue
            else:
                if eis_elem_number == 1:
                    p1.append(parameters[index])

                elif eis_elem_number > 1 and j < eis_elem_number:
                    p1.append(parameters[index])
                    if nleis_elem_number > eis_elem_number:
                        p2.append(parameters[index])
                else:
                    p2.append(parameters[index])

                index += 1

    return p1, p2



import warnings

import numpy as np
from scipy.linalg import inv
from scipy.optimize import curve_fit, basinhopping
from impedance.models.circuits.elements import circuit_elements
from impedance.models.circuits.elements import get_element_from_name
from impedance.models.circuits.fitting import check_and_eval, rmse

import networkx as nx
import re

# Note: a lot of codes are directly adopted from impedance.py.,
# which is designed to be enable a easy integration in the future,
# but now we are keep them separated to ensure the stable performance

ints = '0123456789'


def mae(a, b):
    """
    Calculate the Mean Absolute Error (MAE) between two arrays.

    Parameters
    ----------
    a : numpy.ndarray
        Array of experimental data.

    b : numpy.ndarray
        Array of model fit data.

    Returns
    -------
    float
        The calculated Mean Absolute Error between `a` and `b`.
    """
    return (np.mean(abs(a-b)))


def mape(a, b):
    """
    Calculate the Mean Absolute Percentage Error (MAPE) between two arrays.

    Parameters
    ----------
    a : numpy.ndarray
        Array of experimental data.

    b : numpy.ndarray
        Array of model fit data.

    Returns
    -------
    float
        The calculated Mean Absolute Percentage Error (MAPE)
        between `a` and `b`, expressed as a percentage.
    """

    return (np.mean(abs(a-b)/abs(a))*100)


def seq_fit_param(input_dic, target_arr, output_arr):
    '''
    Convert obtained EIS results to a constant dictionary for
    2nd-NLEIS analysis using sequential optimization as discussed in [1].

    Parameters
    ----------
    input_dic : dict
        Dictionary of EIS fitting results.

    target_arr : list of str
        A list of EIS circuit elements that need to be converted.
        Example: ['TDS0', 'TDS1'].

    output_arr : list of str
        A list of 2nd-NLEIS circuit elements that will be converted to.
        Example: ['TDSn0', 'TDSn1'].

    Raises
    ------
    ValueError
        Raised if the `target_arr` and `output_arr` have different lengths.

    Returns
    -------
    output_dic : dict
        A dictionary of constants that can be used for sequential 2nd-NLEIS
        optimization.

    References
    ----------
    [1] Y. Ji, D.T. Schwartz,
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy:
    II.Model-Based Analysis of Lithium-Ion Battery Experiments.
    J. Electrochem. Soc., 2024. `doi:10.1149/1945-7111/ad2596
    <https://iopscience.iop.org/article/10.1149/1945-7111/ad2596>`_.

    '''
    output_dic = {}
    n = len(target_arr)
    if n != len(output_arr):
        raise ValueError(
            'Target Array and Output Array Must Have the Same Length ')

    for i in range(0, n):
        for j in range(check_and_eval(target_arr[i][0:-1]).num_params):
            if target_arr[i]+'_'+str(j) not in input_dic:
                continue
            output_dic[output_arr[i]+'_' +
                       str(j)] = input_dic[target_arr[i]+'_'+str(j)]
    return (output_dic)


def set_default_bounds(circuit, constants={}):
    """
    Set default bounds for optimization.

    This function assigns default bounds for all parameters in the given
    circuit. The default lower bound is 0, and the upper bound is `np.inf`.

    Exceptions:

    - CPE and La alpha have an upper bound of 1.

    - Symmetry parameter (ε) for 2nd-NLEIS has bounds between -0.5 and 0.5.

    - Curvature parameter (κ) for 2nd-NLEIS has bounds
      between -np.inf and np.inf.

    Parameters
    ----------
    circuit : str
        String defining the equivalent circuit to be fit.

    constants : dict, optional
        Dictionary of parameters and their values to be held constant during
        fitting (e.g., {"R0": 0.1}). Defaults to an empty dictionary.

    Returns
    -------
    bounds : tuple of list
        A 2-tuple containing lists of lower and
        upper bounds for the parameters.

    """

    # extract the elements from the circuit
    extracted_elements = extract_circuit_elements(circuit)

    # loop through bounds
    lower_bounds, upper_bounds = [], []

    for elem in extracted_elements:

        raw_element = get_element_from_name(elem)

        for i in range(check_and_eval(raw_element).num_params):

            if elem in constants or elem + f'_{i}' in constants:

                continue

            elif raw_element in ['CPE', 'La'] and i == 1:
                upper_bounds.append(1)
                lower_bounds.append(0)
            # The following are for nleis.py

            elif raw_element in ['TPn'] and i == 3:
                upper_bounds.append(0.5)
                lower_bounds.append(-0.5)
            elif raw_element in ['RCn'] and i == 2:
                upper_bounds.append(0.5)
                lower_bounds.append(-0.5)
            elif raw_element in ['TDSn', 'TDPn', 'TDCn', 'RCSQn',
                                 'RCDQn'] and (i == 5):
                upper_bounds.append(np.inf)
                lower_bounds.append(-np.inf)
            elif raw_element in ['TDSn', 'TDPn', 'TDCn', 'RCSQn',
                                 'RCDQn'] and i == 6:
                upper_bounds.append(0.5)
                lower_bounds.append(-0.5)
            elif raw_element in ['RCDn', 'RCSn'] and (i == 4):
                upper_bounds.append(np.inf)
                lower_bounds.append(-np.inf)
            elif raw_element in ['RCDn', 'RCSn'] and i == 5:
                upper_bounds.append(0.5)
                lower_bounds.append(-0.5)
            elif raw_element in ['TLMn'] and (i == 6 or i == 7):
                upper_bounds.append(0.5)
                lower_bounds.append(-0.5)
            elif raw_element in ['TLMSn', 'TLMDn'] and (i == 9 or i == 10):
                upper_bounds.append(0.5)
                lower_bounds.append(-0.5)
            elif raw_element in ['TLMSn', 'TLMDn'] and (i == 8):
                upper_bounds.append(np.inf)
                lower_bounds.append(-np.inf)
            elif raw_element in ['RCSQ', 'RCSQn',
                                 'RCDQ', 'RCDQn'] and (i == 2):
                upper_bounds.append(1)
                lower_bounds.append(0)
            else:
                upper_bounds.append(np.inf)
                lower_bounds.append(0)

    bounds = ((lower_bounds), (upper_bounds))

    return bounds

# adopt from impedance.py to support 2nd-NLEIS/NLEIS fitting


def circuit_fit(frequencies, impedances, circuit, initial_guess, constants={},
                bounds=None, weight_by_modulus=False, global_opt=False,
                graph=False,
                **kwargs):
    """ Main function for fitting an equivalent circuit to data.

    By default, this function uses `scipy.optimize.curve_fit
    <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.curve_fit.html>`_
    to fit the equivalent circuit. This function generally works well for
    simple circuits. However, the final results may be sensitive to
    the initial conditions for more complex circuits. In these cases,
    the `scipy.optimize.basinhopping
    <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.basinhopping.html>`_
    global optimization algorithm can be used to attempt a better fit.

    Parameters
    ----------
    frequencies : numpy array
        Frequencies

    impedances : numpy array of dtype 'complex128'
        Impedances

    circuit : string
        String defining the equivalent circuit to be fit

    initial_guess : list of floats
        Initial guesses for the fit parameters

    constants : dictionary, optional
        Parameters and their values to hold constant during fitting
        (e.g. {"RO": 0.1}). Defaults to {}

    bounds : 2-tuple of array_like, optional
        Lower and upper bounds on parameters. Defaults to bounds on all
        parameters of 0 and np.inf, except the CPE alpha
        which has an upper bound of 1

    weight_by_modulus : bool, optional
        Uses the modulus of each data (`|Z|`) as the weighting factor.
        Standard weighting scheme when experimental variances are unavailable.
        Only applicable when global_opt = False

    global_opt : bool, optional
        If global optimization should be used (uses the basinhopping
        algorithm). Defaults to False

    graph : bool, optional
        Whether to use execution graph to process the circuit.
        Defaults to False, which uses eval based code

    kwargs :
        Keyword arguments passed to scipy.optimize.curve_fit or
        scipy.optimize.basinhopping

    Returns
    -------
    p_values : list of floats
        best fit parameters for specified equivalent circuit

    p_errors : list of floats
        one standard deviation error estimates for fit parameters

    Notes
    -----
    Need to do a better job of handling errors in fitting.
    Currently, an error of -1 is returned.

    """

    f = np.array(frequencies, dtype=float)
    Z = np.array(impedances, dtype=complex)

    # set upper and lower bounds on a per-element basis
    if bounds is None:
        bounds = set_default_bounds(circuit, constants=constants)

    cg = CircuitGraph(circuit, constants)

    if not global_opt:
        if 'maxfev' not in kwargs:
            kwargs['maxfev'] = int(1e5)
        if 'ftol' not in kwargs:
            kwargs['ftol'] = 1e-13

        # weighting scheme for fitting
        if weight_by_modulus:
            abs_Z = np.abs(Z)
            kwargs['sigma'] = np.hstack([abs_Z, abs_Z])
        if graph:
            popt, pcov = curve_fit(cg.compute_long, f,
                                   np.hstack([Z.real, Z.imag]),
                                   p0=initial_guess,
                                   bounds=bounds,
                                   **kwargs,
                                   )
        else:
            popt, pcov = curve_fit(wrapCircuit(circuit, constants), f,
                                   np.hstack([Z.real, Z.imag]),
                                   p0=initial_guess, bounds=bounds, **kwargs)

        # Calculate one standard deviation error estimates for fit parameters,
        # defined as the square root of the diagonal of the covariance matrix.
        # https://stackoverflow.com/a/52275674/5144795
        perror = np.sqrt(np.diag(pcov))

    else:
        if 'seed' not in kwargs:
            kwargs['seed'] = 0

        def opt_function(x):
            """ Short function for basinhopping to optimize over.
            We want to minimize the RMSE between the fit and the data.

            Parameters
            ----------
            x : args
                Parameters for optimization.

            Returns
            -------
            function
                Returns a function (RMSE as a function of parameters).
            """
            return rmse(wrapCircuit(circuit, constants)(f, *x),
                        np.hstack([Z.real, Z.imag]))

        def opt_function_graph(x):
            return rmse(cg.compute_long(f, *x), np.hstack([Z.real, Z.imag]))

        class BasinhoppingBounds(object):
            """ Adapted from the basinhopping documetation
            https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.basinhopping.html
            """

            def __init__(self, xmin, xmax):
                self.xmin = np.array(xmin)
                self.xmax = np.array(xmax)

            def __call__(self, **kwargs):
                x = kwargs['x_new']
                tmax = bool(np.all(x <= self.xmax))
                tmin = bool(np.all(x >= self.xmin))
                return tmax and tmin

        basinhopping_bounds = BasinhoppingBounds(xmin=bounds[0],
                                                 xmax=bounds[1])
        if graph:
            results = basinhopping(opt_function_graph, x0=initial_guess,
                                   accept_test=basinhopping_bounds, **kwargs)
        else:
            results = basinhopping(opt_function, x0=initial_guess,
                                   accept_test=basinhopping_bounds, **kwargs)
        popt = results.x

        # Calculate perror
        jac = results.lowest_optimization_result['jac'][np.newaxis]
        try:
            # jacobian -> covariance
            # https://stats.stackexchange.com/q/231868
            pcov = inv(np.dot(jac.T, jac)) * opt_function(popt) ** 2
            # covariance -> perror (one standard deviation
            # error estimates for fit parameters)
            perror = np.sqrt(np.diag(pcov))
        except (ValueError, np.linalg.LinAlgError):
            warnings.warn('Failed to compute perror')
            perror = None

    return popt, perror

# adopt from impedance.py to support 2nd-NLEIS/NLEIS fitting


def wrapCircuit(circuit, constants):
    """ wraps function so we can pass the circuit string """
    def wrappedCircuit(frequencies, *parameters):
        """ returns a stacked array of real and imaginary impedance
        components

        Parameters
        ----------
        circuit : string
        constants : dict
        parameters : list of floats
        frequencies : list of floats

        Returns
        -------
        array of floats

        """

        x = eval(buildCircuit(circuit, frequencies, *parameters,
                              constants=constants, eval_string='',
                              index=0)[0],
                 circuit_elements)
        y_real = np.real(x)
        y_imag = np.imag(x)

        return np.hstack([y_real, y_imag])
    return wrappedCircuit

# modified to enable subtraction


def buildCircuit(circuit, frequencies, *parameters,
                 constants=None, eval_string='', index=0):
    """ recursive function that transforms a circuit, parameters, and
    frequencies into a string that can be evaluated

    Parameters
    ----------
    circuit: str
    frequencies: list/tuple/array of floats
    parameters: list/tuple/array of floats
    constants: dict

    Returns
    -------
    eval_string: str
        Python expression for calculating the resulting fit
    index: int
        Tracks parameter index through recursive calling of the function

    """

    parameters = np.array(parameters).tolist()
    frequencies = np.array(frequencies).tolist()
    circuit = circuit.replace(' ', '')

    def parse_circuit(circuit, parallel=False, series=False, difference=False):
        """ Splits a circuit string by either dashes (series) or commas
            (parallel) outside of any paranthesis. Removes any leading 'p('
            or trailing ')' when in parallel mode or 'd(' or trailing ')'
            when in difference mode. This is adapted from impedance.py to
            support subtraction """

        assert (parallel != series
                or series != difference
                or difference != parallel), \
            'Exactly one of parallel or series or difference must be True'

        def count_parens(string):
            return string.count('('), string.count(')')

        if parallel:
            special = ','
            if circuit.endswith(')') and circuit.startswith('p('):
                circuit = circuit[2:-1]
        if difference:
            special = ','
            if circuit.endswith(')') and circuit.startswith('d('):
                circuit = circuit[2:-1]

        if series:
            special = '-'

        split = circuit.split(special)

        result = []
        skipped = []
        for i, sub_str in enumerate(split):
            if i not in skipped:
                if '(' not in sub_str and ')' not in sub_str:
                    result.append(sub_str)
                else:
                    open_parens, closed_parens = count_parens(sub_str)
                    if open_parens == closed_parens:
                        result.append(sub_str)
                    else:
                        uneven = True
                        while i < len(split) - 1 and uneven:
                            sub_str += special + split[i+1]

                            open_parens, closed_parens = count_parens(sub_str)
                            uneven = open_parens != closed_parens

                            i += 1
                            skipped.append(i)
                        result.append(sub_str)
        return result

    parallel = parse_circuit(circuit, parallel=True)
    series = parse_circuit(circuit, series=True)
    difference = parse_circuit(circuit, difference=True)

    if series is not None and len(series) > 1:
        eval_string += "s(["
        split = series
    elif parallel is not None and len(parallel) > 1:
        eval_string += "p(["
        split = parallel

    # added for nleis.py
    elif difference is not None and len(difference) > 1:
        eval_string += "d(["
        split = difference

    elif series == parallel == difference:  # only single element
        split = series

    for i, elem in enumerate(split):
        if ',' in elem or '-' in elem:
            eval_string, index = buildCircuit(elem, frequencies,
                                              *parameters,
                                              constants=constants,
                                              eval_string=eval_string,
                                              index=index)
        else:
            param_string = ""
            raw_elem = get_element_from_name(elem)
            elem_number = check_and_eval(raw_elem).num_params
            param_list = []
            for j in range(elem_number):
                if elem_number > 1:
                    current_elem = elem + '_{}'.format(j)
                else:
                    current_elem = elem

                if current_elem in constants.keys():
                    param_list.append(constants[current_elem])
                else:
                    param_list.append(parameters[index])
                    index += 1

            param_string += str(param_list)
            new = raw_elem + '(' + param_string + ',' + str(frequencies) + ')'
            eval_string += new

        if i == len(split) - 1:
            if len(split) > 1:  # do not add closing brackets if single element
                eval_string += '])'
        else:
            eval_string += ','

    return eval_string, index

# adopt from impedance.py to support 2nd-NLEIS/NLEIS fitting


def calculateCircuitLength(circuit):
    """ Calculates the number of elements in the circuit.

    Parameters
    ----------
    circuit : str
        Circuit string.

    Returns
    -------
    length : int
        Length of circuit.

    """
    length = 0
    if circuit:
        extracted_elements = extract_circuit_elements(circuit)
        for elem in extracted_elements:
            raw_element = get_element_from_name(elem)
            num_params = check_and_eval(raw_element).num_params
            length += num_params
    return length

# modified to enable subtraction


def extract_circuit_elements(circuit):
    """ Extracts circuit elements from a circuit string.

    Parameters
    ----------
    circuit : str
        Circuit string.

    Returns
    -------
    extracted_elements : list
        list of extracted elements.

    """
    p_string = [x for x in circuit if x not in 'p(),-d']
    extracted_elements = []
    current_element = []
    length = len(p_string)
    for i, char in enumerate(p_string):
        if char not in ints:
            current_element.append(char)
        else:
            # min to prevent looking ahead past end of list
            if p_string[min(i+1, length-1)] not in ints:
                current_element.append(char)
                extracted_elements.append(''.join(current_element))
                current_element = []
            else:
                current_element.append(char)
    extracted_elements.append(''.join(current_element))
    return extracted_elements

# Circuit Graph for computation optimization
# Special Thanks to Jake Anderson for the original code


class CircuitGraph:
    '''
    A class to represent a circuit as a directed graph.
    '''
    # regular expression to find parallel and difference blocks
    _parallel_difference_block_expression = re.compile(r'(?:p|d)\([^()]*\)')

    # regular expression to remove whitespace
    _whitespce = re.compile(r"\s+")

    def __init__(self, circuit, constants=None):
        '''
        Initialize the CircuitGraph object.'''
        # remove all whitespace from the circuit string
        self.circuit = self._whitespce.sub("", circuit)
        # parse the circuit string and initialize the graph
        self.parse_circuit()
        # compute the execution order of the graph
        self.execution_order = list(nx.topological_sort(self.graph))
        # initialize the constants dictionary
        self.constants = constants if constants is not None else dict()

    def parse_circuit(self):
        '''
        Parse the circuit string and initialize the graph.
        '''
        # initialize the node counters for each type of block
        self.snum = 1
        self.pnum = 1
        self.dnum = 1
        # initialize the circuit string to be parsed
        parsing_circuit = self.circuit

        # determine all of the base elements, their functions
        # and add them to the graph
        element_name = extract_circuit_elements(parsing_circuit)
        element_func = [
            circuit_elements[get_element_from_name(e)] for e in element_name
        ]
        # graph initialization
        self.graph = nx.DiGraph()
        # add nodes to the graph
        for e, f in zip(element_name, element_func):
            self.graph.add_node(e, Z=f)

        # find unnested parallel and difference blocks
        pd_blocks = self._parallel_difference_block_expression.findall(
            parsing_circuit)

        while len(pd_blocks) > 0:
            # add parallel or difference blocks to the graph
            # unnesting each time around the loop
            for pd in pd_blocks:
                operator = pd[0]
                pd_elem = pd[2:-1].split(",")

                if operator == "p":
                    nnum = self.pnum
                    self.pnum += 1
                elif operator == "d":
                    nnum = self.dnum
                    self.dnum += 1

                node = f"{operator}{nnum}"
                self.graph.add_node(node, Z=circuit_elements[operator])
                for elem in pd_elem:
                    elem = self.add_series_elements(elem)
                    self.graph.add_edge(elem, node)
                parsing_circuit = parsing_circuit.replace(pd, node)

            pd_blocks = self._parallel_difference_block_expression.findall(
                parsing_circuit)

        # pick up any top line series connections
        self.add_series_elements(parsing_circuit)

        # assign layers to the nodes
        for layer, nodes in enumerate(nx.topological_generations(self.graph)):
            for n in nodes:
                self.graph.nodes[n]["layer"] = layer
    # function to add series elements to the graph

    def add_series_elements(self, elem):
        '''
        Add series elements to the graph.
        '''
        selem = elem.split("-")
        if len(selem) > 1:
            node = f"s{self.snum}"
            self.snum += 1
            self.graph.add_node(node, Z=circuit_elements["s"])
            for n in selem:
                self.graph.add_edge(n, node)
            return node

        # if there isn't a series connection in elem just return it unchanged
        return selem[0]

    # function to visualize the graph
    def visualize_graph(self, **kwargs):
        '''
        Visualize the graph.'''
        pos = nx.multipartite_layout(self.graph, subset_key="layer")
        nx.draw_networkx(self.graph, pos=pos, **kwargs)

    # function to compute the impedance of the circuit
    def compute(self, f, *parameters):
        '''
        Compute the impedance of the circuit at the given frequencies.
        '''
        node_results = {}
        pindex = 0
        for node in self.execution_order:
            Zfunc = self.graph.nodes[node]["Z"]
            plist = [
                node_results[pred] for pred in self.graph.predecessors(node)
            ]

            if len(plist) < 1:
                n_params = Zfunc.num_params
                for j in range(n_params):
                    p_name = format_parameter_name(node, j, n_params)
                    if p_name in self.constants:
                        plist.append(self.constants[p_name])
                    else:
                        plist.append(parameters[pindex])
                        pindex += 1
                node_results[node] = Zfunc(plist, f)
            else:
                node_results[node] = Zfunc(plist)

        return np.squeeze(node_results[node])

    # To enable comparision

    def __eq__(self, other):
        '''
        Compare two CircuitGraph objects for equality.
        '''
        if not isinstance(other, CircuitGraph):
            return False
        # Compare the internal graph attributes
        return (self.graph.nodes == other.graph.nodes
                and self.graph.edges == other.graph.edges)

    # To enable direct calling

    def __call__(self, f, *parameters):
        '''
        Compute the impedance of the circuit at the given frequencies.
        '''
        Z = self.compute(f, *parameters)
        return Z

    def compute_long(self, f, *parameters):
        '''
        Compute the impedance of the circuit at the given frequencies.
        And convert it to a long array for curve_fit.
        '''
        Z = self.compute(f, *parameters)
        return np.hstack([Z.real, Z.imag])

    def calculate_circuit_length(self):
        '''
        calculate the number of parameters in the circuit
        '''
        n_params = [
            getattr(Zfunc, "num_params", 0)
            for node, Zfunc in self.graph.nodes(data="Z")
        ]
        return np.sum(n_params)


def format_parameter_name(name, j, n_params):
    '''
    Format the parameter name for the given element.
    '''
    return f"{name}_{j}" if n_params > 1 else f"{name}"



import matplotlib.pyplot as plt
import numpy as np
from nleis.visualization import plot_altair, plot_first, plot_second
import json


def test_plot_nyquist():

    Z = np.array([1, 2, 3]) + 1j*np.array([2, 3, 4])

    # pass axes
    _, ax = plt.subplots()
    ax = plot_first(ax, Z, scale=10)
    xs, ys = ax.lines[0].get_xydata().T
    assert (xs == Z.real).all() and (ys == -Z.imag).all()

    # pass axes
    _, ax = plt.subplots()
    ax = plot_second(ax, Z, scale=10)
    xs, ys = ax.lines[0].get_xydata().T
    assert (xs == Z.real).all() and (ys == -Z.imag).all()


def test_plot_altair():
    frequencies = [1000.0, 1.0, 0.01]
    Z = np.array([1, 2, 3]) + 1j*np.array([2, 3, 4])

    chart = plot_altair({'data': {'f': frequencies, 'Z': Z},
                         'fit': {'f': frequencies, 'Z': Z, 'fmt': '-'}},
                        size=400)

    datasets = json.loads(chart.to_json())['datasets']
    for dataset in datasets.keys():
        assert len(datasets[dataset]) == len(Z)
